"""
This module contains various functions to infer parameters of archpy model. For now, only surface parameters can be inferred.
"""

import numpy as np
import matplotlib.pyplot as plt
import geone
import geone.covModel as gcm
import sys
import ipywidgets as widgets
from ipywidgets import interact, fixed, Button

#my modules
import ArchPy

# functions
def cm_any_nan(cm):
    
    """
    Detect if there is a nan inside a covmodel

    Parameters
    ----------
    cm: :class:`geone.covModel.CovModel1D`, :class:`geone.covModel.CovModel2D` or :class:`geone.covModel.CovModel3D`
        covmodel to check
    """
    
    yes = False

    if isinstance(cm, geone.covModel.CovModel2D):
        dim = 2
    if isinstance(cm, geone.covModel.CovModel1D):
        dim = 1
    if isinstance(cm, geone.covModel.CovModel3D):
        dim = 3

    elems = [i[1] for i in cm.elem]
    for el in elems:
        for k,v in el.items():
            if k in ("w", "nu", "s"):
                if np.isnan(v):
                    yes = True
                    break
            if k == "r":
                if dim > 1:
                    for ri in v:
                        if np.isnan(ri):
                            yes = True
                            break
                            
                            
def infer_surface(ArchTable, unit, hmax=np.nan, cm_to_fit=None, auto=True, dim=1, plot=True,
                  npoints_min=20, max_nugget=1, bounds = None, default_covmodel=None, vb=1, **kwargs):
    
    """
    Infer surface parameters of a unit in an :class:`base.Arch_table`. Fitting can be done automatically or manually.

    Parameters
    ----------
    ArchTable: :class:`base.Arch_table`
        ArchTable to infer surface parameters
    unit: :class:`base.Unit`
        unit to infer surface parameters
    hmax: float, optional  
        maximum distance to consider for variogram inference
    cm_to_fit: :class:`geone.covModel.CovModel1D`, :class:`geone.covModel.CovModel2D` or :class:`geone.covModel.CovModel3D`, optional
        covmodel to fit
    auto: bool, optional
        if True, automatic inference of the covmodel
    dim: int, optional
        dimension of the covmodel to fit
    plot: bool, optional
        if True, plot the variogram
    npoints_min: int, optional
        minimum number of points to infer the covmodel
    max_nugget: float, optional
        maximum nugget effect to infer
    bounds: list, optional  
        bounds for the parameters of the covmodel to fit
        list containing two lists, one for the lower bounds and one for the upper bounds of each parameter

    Returns
    -------
    cm: :class:`geone.covModel.CovModel1D`, :class:`geone.covModel.CovModel2D` or :class:`geone.covModel.CovModel3D`
        inferred covmodel
    """
    
    if plot:
        plt.cla()
        plt.close()
        global fig, ax  # global figure to allow class to access to the plot
        fig,ax=plt.subplots()
        
    surf = unit.surface

    x = np.array([surf.x, surf.y]).T
    v = np.array(surf.z)
    Lz = ArchTable.get_zg()[-1] - ArchTable.get_zg()[0]
    Lx = ArchTable.get_xg()[-1] - ArchTable.get_xg()[0]
    Ly = ArchTable.get_yg()[-1] - ArchTable.get_yg()[0]
    
    dmax = np.sqrt(Lx**2 + Ly**2)

    if len(surf.z) > npoints_min:
        
        if auto:  # automatic inference
            
            if dim == 1:
                
                # compute var exp
                h,exp_var, p = gcm.variogramExp1D(x, v, hmax=hmax, ncla=20, make_plot = False, **kwargs)  # just make a plot of the exp variogram

                if plot:
                    plt.scatter(h, exp_var)
                
                # boundaries
                vmax = np.var(v)*2
                rmax = dmax*2                
                
                # covmodel to fit
                if cm_to_fit is None:
                    if cm_any_nan(surf.covmodel):  # check if there is any nan inside cm
                        cm_to_fit = surf.covmodel  # if so it will be passed to fit function
                    else:
                        # default covmodel to fit
                        cm_to_fit = gcm.CovModel1D(elem=[("cubic",{"w":np.nan,"r":np.nan}),
                                                         ("exponential",{"w":np.nan,"r":np.nan}),
                                                         ("spherical",{"w":np.nan,"r":np.nan}),
                                                         ("nugget",{"w":np.nan})])
                        
                        bounds = ((0, 0, 0, 0, 0, 0, 0),  # min bounds
                                  (np.var(v), dmax*2, np.var(v), dmax*2, np.var(v), dmax*2, max_nugget)) # max bounds
                     
                else:
                    # ensure that bounds exist
                    if bounds is None:
                        bmin = []
                        bmax = []

                        for el in cm_to_fit.elem:
                            if el[0] is not "nugget":
                                for k in el[1].keys():
                                    if k == "w":
                                        bmin.append(0)
                                        bmax.append(vmax)
                                    elif k == "r":
                                        bmin.append(0)
                                        bmax.append(rmax)
                            else:
                                bmin.append(0)
                                bmax.append(max_nugget)
                        bounds = (bmin, bmax)
                
                # automatic fitting
                cm_fitted = gcm.covModel1D_fit(x, v, cm_to_fit, 
                                               hmax=hmax, bounds=bounds, make_plot=False)[0]
                
                
                #TO DO : add a check to see if fitting is good enough

                #remove unutilized structures
                l = cm_fitted.elem.copy()
                sum_w = 0
                for e in l:
                    if e[1]["w"] > sum_w:
                        sum_w += e[1]["w"]

                for e in l:
                    if (e[1]["w"]/sum_w) < 0.01:
                        cm_fitted.elem.remove(e)    

                # finalization                              
                surf.set_covmodel(gcm.covModel1D_to_covModel2D(cm_fitted))
                if plot:
                    plt.title(unit.name)
                    cm_fitted.plot_model(vario=True, c="k")
                    if ~np.isnan(hmax):
                        plt.xlim(-hmax*0.1, hmax)
                    plt.xlabel("h [L]")
                    plt.legend() 
                    plt.show() 
                

            elif dim == 2: 
                
                # boundaries
                vmax = np.var(v)*2
                rmax = dmax*2  
                
                # covmodel to fit
                if cm_to_fit is None:
                    if cm_any_nan(surf.covmodel):  # check if there is any nan inside cm
                        cm_to_fit = surf.covmodel  # if so it will be passed to fit function
                    else:
                        # default covmodel to fit
                        cm_to_fit = gcm.CovModel2D(elem=[("cubic",{"w":np.nan,"r":[np.nan, np.nan]}),
                                                         ("exponential",{"w":np.nan,"r":[np.nan, np.nan]}),
                                                         ("spherical",{"w":np.nan,"r":[np.nan, np.nan]}),
                                                         ("nugget",{"w":np.nan})], alpha=np.nan)
                        
                        bounds = ((0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -45),  # min bounds
                                  (vmax, rmax, rmax,
                                   vmax, rmax, rmax,
                                   vmax, rmax, rmax,
                                   max_nugget,
                                   45)) # max bounds
                     
                else:
                    # ensure that bounds exist
                    if bounds is None:
                        bmin = []
                        bmax = []

                        for el in cm_to_fit.elem:
                            if el[0] is not "nugget":
                                for k in el[1].keys():
                                    if k == "w":
                                        bmin.append(0)
                                        bmax.append(vmax)
                                    elif k == "r":
                                        for o in range(2):
                                            bmin.append(0)
                                            bmax.append(rmax)
                            else:
                                bmin.append(0)
                                bmax.append(max_nugget)
                        
                        # alpha
                        bmin.append(-45)
                        bmax.append(45)
                        bounds = (bmin, bmax)
                
                # automatic fitting
                cm_fitted = gcm.covModel2D_fit(x, v, cm_to_fit, 
                                               hmax=hmax, bounds=bounds, make_plot=False)[0]
                
                (h1, v1, p1), (h2, v2, p2) = gcm.variogramExp2D(x, v, cm_fitted.alpha, hmax=(hmax, hmax), 
                                                                make_plot=False, **kwargs)
                
                if plot:
                    plt.scatter(h1, v1, color="lightblue", label="x : var-exp")
                    plt.scatter(h2, v2, color="orange", label="y : var-exp")
                    
                    plt.title(r"{} : $\alpha$ {}".format(unit.name, np.round(cm_fitted.alpha, 2)))
                    
                    cm_fitted.plot_model_one_curve(1, 1, hmax=2*dmax, c="lightblue", label= "x' model")
                    cm_fitted.plot_model_one_curve(2, 1, hmax=2*dmax, c="orange", label="y' model")
                
                    if ~np.isnan(hmax):
                        plt.xlim(-hmax*0.1, hmax)
                    plt.xlabel("h [L]")
                    plt.legend()  
                    
                # set model
                surf.set_covmodel(cm_fitted)
                
        else : # manual fitting
            
            #create a continue Button
            button = Button(description="Continue")  # continue button
            output = widgets.Output()
            display(button, output)
            
            # function to separate exp var and fitting
            def on_button_clicked(b):
                global i
                with output:
                    if i == 0:
                        plt.grid()
                        dim = ev.dim  # which dimension ? 
                        hmax = ev.hmax
                        cm_to_fit.h_max = np.max(hmax)
                        cm_to_fit.alpha = ev.alpha
                        ev.clear()
                        cm_to_fit.fit(dim=dim)

                        plt.xlim(0, right = 1.1*np.nanmax(ev.h))
                        plt.ylim(0, top = 1.1*np.nanmax(ev.val))
                    elif i > 0:
                        print("Finished and variogram saved")
                        cm_to_fit.clear()
                        button.close()
                        if isinstance(cm_to_fit.cm, gcm.CovModel1D):
                            surf.set_covmodel(gcm.covModel1D_to_covModel2D(cm_to_fit.cm))
                        else:
                            surf.set_covmodel(cm_to_fit.cm)
                    i+=1
            
            
            ev = Var_exp(x, v, hmax_lim=dmax*2)  # object for experimental variogram setting
            ev.fit()
            if np.isnan(hmax):
                cm_to_fit = Cm2fit(h_max=np.max((Lx*2, Ly*2, Lz*2)), r_max=2*np.max((Lx*2, Ly*2, Lz*2)), w_max=2*np.var(v))  # object for fitting a covmodel object (see geone doc.)
            else:
                cm_to_fit = Cm2fit(h_max=hmax, r_max=2*np.max((Lx*2, Ly*2, Lz*2)), w_max=2*np.var(v))  # object for fitting a covmodel object (see geone doc.)
            global i
            i = 0
            button.on_click(on_button_clicked)
            
            # display(button)  # Important !! --> you have to use display to show the button
            
    else :  # not enough data for a variogram

        if vb:
            print("Not enough data points")

        if not auto:
            def on_button_clicked(b):
                # create default

                if default_covmodel is not None:
                    surf.set_covmodel(default_covmodel)
                else:
                    cm_def = gcm.CovModel2D(elem=[("exponential", {"w":(Lz/4)**2/(128), "r":[min(Lx, Ly)/4, min(Lx,Ly)/4]})])
                    surf.set_covmodel(cm_def)

            button = Button(description="add default covmodel")  # continue button
            display(button)
            button.on_click(on_button_clicked)

        else:
            if vb:
                print("Default covmodel added")
            if default_covmodel is not None:
                surf.set_covmodel(default_covmodel)
            else:
                if surf.covmodel is None:
                    cm_def = gcm.CovModel2D(elem=[("exponential", {"w":(Lz/4)**2/(128), "r":[min(Lx, Ly)/4, min(Lx,Ly)/4]})])
                    surf.set_covmodel(cm_def)
                else:
                    if vb:
                        print("Nothing has changed")

    
                                               
def fit_surfaces(self, default_covmodel=None, **kwargs):

    """
    Function to fit a covmodel to each surface of an :class:`Arch_table` object

    Parameters
    ----------
    self : :class:`base.Arch_table` object
        Arch_table object to fit
    default_covmodel : :class:`geone.CovModel2D` object, optional
        Default covmodel to add if not enough data for a variogram
    **kwargs :
        Arguments for :func:`infer_surface`. See :func:`infer_surface` for more details
    """

    print(default_covmodel)
    # create a nested function to set default argument (ArchTable and surface)
    def f(unit, auto, default_covmodel, **kwargs):
        unit = self.get_unit(unit)
        def _f():
            infer_surface(self, unit, auto=auto, default_covmodel=default_covmodel, **kwargs)
        widgets.interact(_f, **kwargs)
        
    units_n = [i.name for i in self.get_all_units()]

    widgets.interact(f, unit=units_n, auto=False, dim=[1, 2], default_covmodel=fixed(default_covmodel), **kwargs)
    


# Classes

# var exp
## Class for estimate var exp
def plot_var_exp(h1, h2, v1 ,v2, p1, p2, print_pairs=False):
    
    """
    Function to plot experimental variogram

    Parameters
    ----------
    h1 : array
        spacing lag between points in x direction
    h2 : array
        spacing lag between points in y direction
    v1 : array
        variance between points in x direction
    v2 : array
        variance between points in y direction
    p1 : array
        number of pairs between points in x direction
    p2 : array
        number of pairs between points in y direction
    print_pairs : bool, optional
        print number of pairs on the plot. The default is False.
    """
    
    plt.scatter(h1, v1, label="x'")
    plt.scatter(h2, v2, label="y'")
    plt.plot(h1, v1)
    plt.plot(h2, v2)   
    
    if print_pairs:
        for i in range(len(h1)):
            plt.text(h1[i],v1[i],p1[i]) 
        for i in range(len(h2)):
            plt.text(h2[i], v2[i], p2[i]) 
    
    plt.legend()
    plt.grid()
    plt.show()
    
class Var_exp():
    
    """
    Class to estimate experimental variogram

    Parameters
    ----------
    x : array
        x coordinates of the points. size = (n, dim) with dim = 1, 2 or 3
    v : array
        values
    hmax_lim : float
        maximum distance to consider for the variogram
    ax : matplotlib axis
        axis to plot the variogram
    """

    def __init__(self, x, v, hmax_lim=1000, ax=None):
        
        self.x = np.array(x)
        self.v = np.array(v)
        self.h = None # results
        self.val = None # results
        self.hmax_lim = hmax_lim
        self.dim = 1
        self.hmax = 0
        self.alpha = 0
        self.ax = ax
        
    # fit function
    def fit(self, **kwargs):
        
        """
        Function to fit the experimental variogram

        Parameters
        ----------
        kwargs : dict
            dictionnary of parameters to pass to make_exp_var
        """

        if self.ax is None:
            [l.remove() for l in ax.lines]
        else:
            [l.remove() for l in self.ax.lines]

        print("General parameters - Do not touch except dim if you want to specify the dimension")
        w=widgets.interact(self.make_exp_var, dim=(1, 3), **kwargs)
        self.w = w
    
    def clear(self):

        """
        :meta_private:
        """
        self.w.widget.close()
    
    
    # functions
    def make_exp_var(self, dim=1, **kwargs):

        """
        Function to estimate the experimental variogram

        Parameters
        ----------
        dim : int
            dimension of the problem
        kwargs : dict
            dictionnary of parameters to pass to make_exp_var_1D, make_exp_var_2D or make_exp_var_3D
        """

        print("Experimental variogram parameters")
        x = self.x
        v = self.v
        self.dim = dim  # store dimension
        if dim == 1:
            widgets.interact(self.make_exp_var_1D, x=fixed(x), v=fixed(v),
                             ncla=(1, 200, 5), hmax=(0, self.hmax_lim, 1), **kwargs)
            
        elif dim == 2:
            widgets.interact(self.make_exp_var_2D, x=fixed(x), v=fixed(v),
                     ncla_x=(1, 50, 1), ncla_y=(1, 50, 1), 
                     hmax_x=(0, self.hmax_lim, 1), hmax_y=(0, self.hmax_lim, 1), 
                     alpha=(-90, 90, 5),
                     make_plot=fixed(False), **kwargs)
            
        elif dim == 3:
            # TO DO
            pass

    def make_exp_var_1D(self, x, v, hmax, ncla=10, **kwargs):

        """
        Function to estimate the experimental variogram in 1D

        Parameters
        ----------

        x : array
            x coordinates of the points. size = (n, dim) with dim = 1, 2 or 3
        v : array
            values
        hmax : float
            maximum distance to consider for the variogram
        ncla : int
            number of classes to consider
        kwargs : dict
            dictionnary of parameters to pass to geone.covModel.variogramExp1D
        """

        plt.cla()
        
        if hmax <= 0:
            hmax = np.nan

        h,v,p = gcm.variogramExp1D(x, v, hmax=hmax, ncla=ncla, make_plot=False, **kwargs)
        
        self.h = h
        self.val = v
        self.hmax = hmax
        plt.scatter(h, v, label="var_exp")
        plt.plot(h, v)
        #plt.bar(h, p, alpha=0.3, width=max(h)/len(h))
        plt.legend()
        plt.grid()
        
    def make_exp_var_2D(self, x, v, hmax_x, hmax_y, ncla_x=10, ncla_y=10, alpha=0, **kwargs):
        
        """
        Function to estimate the experimental variogram in 2D

        Parameters
        ----------
        x : array
            x coordinates of the points. size = (n, dim) with dim = 1, 2 or 3
        v : array
            values
        hmax_x : float
            maximum distance to consider for the variogram in x direction
        hmax_y : float
            maximum distance to consider for the variogram in y direction
        ncla_x : int
            number of classes to consider in x direction
        ncla_y : int   
            number of classes to consider in y direction
        alpha : float
            angle of the direction of the variogram
        kwargs : dict
            dictionnary of parameters to pass to geone.covModel.variogramExp2D
        """

        plt.cla()
        
        if hmax_x <= 0:
            hmax_x = np.nan
        if hmax_y <= 0:
            hmax_y = np.nan
            
        ncla = (ncla_x, ncla_y)
        hmax=(hmax_x, hmax_y)
        self.hmax = hmax
        (h1, v1, p1), (h2, v2, p2) = gcm.variogramExp2D(x, v, alpha, 
                                                        hmax=hmax, ncla=ncla,
                                                        **kwargs)
        
        
        self.h = np.array((h1, h2))
        self.val = np.array((v1, v2))
        self.alpha = alpha
        plot_var_exp(h1, h2, v1 ,v2, p1, p2)
        plt.legend()
        
    def make_exp_var_3D(self, x, v, ncla=(10, 10, 10), hmax=None, alpha=0, beta=0, gamma=0, **kwargs):
        # TO DO
        (h1, v1, p1), (h2, v2, p2), (h3, v3, p3) = gcm.variogramExp3D(x, v,
                                                                      alpha, beta, gamma,
                                                                      hmax=hmax, ncla=ncla,
                                                                      **kwargs)


class Cm2fit():
    
    """
    Class for adjusting variograms

    Parameters
    ----------
    h_max : float
        maximum distance for the variogram
    w_max : float
        maximum variance for the variogram
    r_max : float
        maximum range for the variogram
    nu_max : float
        maximum smoothness (nu parameter) to consider for the variogram, only for Matern covariance
    alpha : float
        angle of the direction of the variogram, 0 implies that x axis os oriented West-East and y axis South-North
        This is just initial value for the fit. Only used for 2D variograms
    ax : matplotlib.axes, optional
        axes to plot the variogram
    """


    # 1D estimation
    def __init__(self, h_max=10, w_max=10, r_max = 100, nu_max=10, alpha=0, ax=None):
        
        self.h_max = h_max
        self.w_max = w_max
        self.r_max = r_max
        self.nu_max = nu_max
        self.alpha = alpha
        self.l = []
        self.w = []
        self.ax = ax

    # update functions 
    def update_hmax(self, h_max):
        
        """
        update the maximum distance for the variogram

        :meta private:
        """
        #remove
        if self.ax is None:
            [l.remove() for l in ax.lines]
        else:
            [l.remove() for l in self.ax.lines]
        #update params
        self.h_max = h_max
        
        # plot
        h = np.arange(0, self.h_max, 0.1)
        f = self.cm.vario_func()
        plt.plot(h, f(h), c="k")
        plt.show()
    
    def update_cm2D(self, h_max, alpha):
        
        """
        Update the maximum distance for the variogram and the angle of the direction of the variogram

        :meta private:
        """

        #remove
        if self.ax is None:
            [l.remove() for l in ax.lines]
        else:
            [l.remove() for l in self.ax.lines]
        
        #update params
        self.h_max = h_max
        self.cm.alpha = alpha
        
        # plot
        self.cm.plot_model_one_curve(1, True, hmax=self.h_max, color="blue")
        self.cm.plot_model_one_curve(2, True, hmax=self.h_max, color="orange")  
        
    
    def make_update(self):
        
        """
        Private function to update the variogram

        :meta private:
        """

        def update(i, typ, **kwargs):
            
            if self.ax is None:
                [l.remove() for l in ax.lines]
                ax.texts.clear()
            else:
                [l.remove() for l in self.ax.lines]
                self.ax.texts.clear()
            
            if typ == "nugget":
                dic = {"w": kwargs["w"]}
            elif typ == "matern":
                dic = {"w": kwargs["w"], "r":kwargs["r"], "nu":kwargs["nu"]}
            elif typ in ("gamma", "power", "exponential_generalized"):
                pass
                # dic = {"w": kwargs["w"], "r":kwargs["r"], "s":kwargs["s"]}
            else:
                dic = {"w": kwargs["w"], "r":kwargs["r"]}
            self.cm.elem[i] = (typ, dic)
            
            f = self.cm.vario_func()
            h = np.arange(0, self.h_max, 0.1)
            plt.plot(h, f(h), c="k")
            
            
        return update

    def make_update_2D(self):

        """
        Private function to update the variogram

        :meta private:
        """

        def update(i, typ, **kwargs):
            
            # remove previous lines
            if self.ax is None:
                for o in range(2):
                    [l.remove() for l in ax.lines]
                ax.texts.clear()
            else:
                for o in range(2):
                    [l.remove() for l in self.ax.lines]
                self.ax.texts.clear()
            
            if typ == "nugget":
                dic = {"w": kwargs["w"]}
            elif typ == "matern":
                dic = {"w": kwargs["w"], "r":[kwargs["rx"], kwargs["ry"]], "nu":kwargs["nu"]}
            elif typ in ("gamma", "power", "exponential_generalized"):
                print("not implemented yet")
                pass
                # dic = {"w": kwargs["w"], "r":kwargs["r"], "s":kwargs["s"]}
            else:
                dic = {"w": kwargs["w"], "r":[kwargs["rx"], kwargs["ry"]]}
            self.cm.elem[i] = (typ, dic)  # change elements
            
            self.cm.plot_model_one_curve(1, True, hmax=self.h_max, color="blue", label="x'")
            self.cm.plot_model_one_curve(2, True, hmax=self.h_max, color="orange", label="y'")
            plt.legend()
        return update
    
    # display widgets
    def choose_struc(self, n_struc=1, dim=1):
        
        """
        Methods to add widgets depending on the numbers of structures to consider

        :meta private:

        Parameters
        ----------
        n_struc : int
            number of structures to consider
        dim : int
            dimension of the problem
        """

        if dim == 1:
            #empty covmodel
            elem = []
            for i in range(n_struc):
                elem.append(("spherical", {"w":1, "r":1}))
            self.cm = gcm.CovModel1D(elem = elem)

            update = self.make_update()
            widgets.interact(self.update_hmax, h_max=(0, 2*self.h_max))
            for i in range(n_struc):
                update = self.make_update()


                wid = widgets.interact(update, i=fixed(i), typ=["spherical", "exponential", "cubic", "gaussian", "matern", "nugget"],
                                          r=(0, self.r_max, self.r_max/1000),
                                          w=(0, self.w_max, self.w_max/1000),
                                          nu=(0, self.nu_max, 0.1))
                self.w.append(wid)
        
        elif dim == 2:  #  in 2D
            elem = []
            for i in range(n_struc):
                elem.append(("spherical", {"w":1, "r":[1, 1]}))
            self.cm = gcm.CovModel2D(elem = elem, alpha=self.alpha)
            
            update = self.make_update_2D()
            widgets.interact(self.update_cm2D, h_max=(0, 2*self.h_max), alpha=fixed(self.alpha))  
            for i in range(n_struc):
                update = self.make_update_2D()
                wid = widgets.interact(update, i=fixed(i), typ=["spherical", "exponential", "cubic","gaussian", "matern", "nugget"],
                                          rx=(0, self.r_max, self.r_max/1000),
                                          ry=(0, self.r_max, self.r_max/1000),
                                          w=(0, self.w_max, self.w_max/1000),
                                          nu=(0, self.nu_max, 0.1))
                self.w.append(wid)

    def fit(self, dim=1):
        
        """
        :meta private:
        """

        plt.grid()
        if self.ax is None:
            [l.remove() for l in ax.lines]
            ax.texts.clear()
        else:
            [l.remove() for l in self.ax.lines]
            self.ax.texts.clear()
        
        wid = widgets.interact(self.choose_struc, n_struc=(0, 10), dim=fixed(dim))
        self.w.append(wid)
        
    def clear(self):

        """
        :meta private:
        """

        for w in self.w:
            w.widget.close()
        


"""
This module contains functions to load and save ArchPy models and results.
"""

import os
import sys
import numpy as np
import matplotlib
from matplotlib.widgets import PolygonSelector, RectangleSelector
import matplotlib.pyplot as plt
import pandas as pd
import copy
import pickle
import yaml

#geone
import geone
import geone.covModel as gcm

#ArchPy
import ArchPy


###utils_inputs####
class drawPoly:
    
    """
    Class to draw a polygon on a specified figure (ax) and return the coordinates of the polygon.
    First the object must be create and after the draw function can be called.
    It takes as inputs a figure (ax) and a type of drawing (rectangle or polygon)
    """

    def __init__(self):
        
        self.list_p=[]
    
    def draw(self, ax, typ="rectangle"):

        """
        Draw a polygon on a specified figure (ax) and return the coordinates of the polygon.
        It takes as inputs a figure (ax) and a type of drawing (rectangle or polygon)

        Parameters
        ----------
        ax : matplotlib axes object
            axes on which the polygon will be drawn
        typ : str, optional
            type of drawing, by default "rectangle".
            Can be "rectangle" or "polygon".
        """

        print("Select your region")
        print("Use left button to select you area(s)")
        if typ=="rectangle":
            self.poly=RectangleSelector(ax, self.f_rect, interactive=True)
        elif typ=="polygon":
            self.poly=PolygonSelector(ax, self.f)
            
    def f_rect(self, eclick, erelease):
        
        self.list_p=[] #remove previous coordinates if any to ensure to have only one rectangle
        x1, y1=eclick.xdata, eclick.ydata
        x2, y2=erelease.xdata, erelease.ydata      
        self.list_p.append( [(x1, y1),(x1, y2),(x2, y2),(x2, y1)])
        
        return 1
    
    def f(self, x):

        if x not in self.list_p:
            self.list_p.append(x)
            print("Polygon added - please press esc if you want to draw another or close the window when you have finished")
        
        return 1


#### Load files ####
def imp_cm(covmodel_dic):
    
    """
    Function to import a geone covmodel given the dictionnary written in the yaml file

    Note
    ----
    The dictionnary must contain the following keys:
    - elem : list of tuples containing the elementary structures
    - alpha : list of floats, anisotropy ratios
    - beta : list of floats, anisotropy ratios
    - gamma : list of floats, anisotropy ratios

    Warning
    -------
    Does not work if the variogram is a pure nugget

    Parameters
    ----------
    covmodel_dic : dict
        dictionnary containing the covmodel parameters
        
    Returns
    -------
    geone.covModel
    """

    # check dimension
    n=len(covmodel_dic["elem"][0][1]["r"])
    if n == 1:
        cm=gcm.CovModel1D(elem=covmodel_dic["elem"])
    elif n == 2:
        cm=gcm.CovModel2D(elem=covmodel_dic["elem"], alpha=covmodel_dic["alpha"])
    elif n == 3:
        cm=gcm.CovModel3D(elem=covmodel_dic["elem"],
                       alpha=covmodel_dic["alpha"],
                       beta=covmodel_dic["beta"],
                       gamma=covmodel_dic["gamma"])
    
    return cm

       
def load_results(ArchTable, surfs=None, surfs_bot=None, units=None, facies=None, props=None):
    
    """
    Load ArchTable results. 
    If external results must be imported, path of these files (e.g. surf argument) can be passed directly
    All files must be pickle binary files with appropriate format (see below).

    Parameters
    ----------
    ArchTable : :class:`base.Arch_table`
                ArchTable object
    surf      : str,
                Path to the surface results file
                The surfaces original format must be a dictionary with 2D arrays containing surfaces
                as values and associated pile objects as keys.
    surf_bot  : str,
                Path to the surface bot results file
                The surfaces original format must be a dictionary with 2D arrays containing surfaces
                as values and associated pile objects as keys.
    units     : str,
                Path to the units results file
                The units original format must be a 4D array of size (nreal, nz, ny, nx)
                with unit IDs as values.
    facies    : str,
                Path to the facies results file
                The facies original format must be a 5D array of size
                (nreal_units, nreal_facies, nz, ny, nx) with facies IDs as values.
    props     : str,
                Path to the properties results file
                The properties original format must be a dictionary with property names as keys and
                6D array of size (nreal_units, nreal_facies, nreal_prop, nz, ny, nx)
                with prop values as values.

    Returns
    -------
    ArchTable : :class:`base.Arch_table`
                ArchTable object with results loaded
    """
       
    if surfs is None:
        #surfaces
        try:
            with open(os.path.join(ArchTable.ws, ArchTable.name + ".sf"), "rb") as f:
                ArchTable.Geol.surfaces_by_piles=pickle.load(f)
                ArchTable.surfaces_computed=1
        except:
            print("Surface results file not found")
    if surfs_bot is None:
        try:
            with open(os.path.join(ArchTable.ws, ArchTable.name + ".sfb"), "rb") as f:
                ArchTable.Geol.surfaces_bot_by_piles=pickle.load(f)
            print("#### Surfaces loaded ####")
            ArchTable.surfaces_computed=1
        except:
            print("Surface bot results file not found")
            
    if ~ArchTable.write_results:
        if units is None:
            #units
            try:
                with open(os.path.join(ArchTable.ws, ArchTable.name + ".unt"), "rb") as f:
                    ArchTable.Geol.units_domains=pickle.load(f)
                print("#### Units loaded ####")
                ArchTable.surfaces_computed=1
            except: 
                print("Unit results file not found")
        if facies is None:
            #facies
            try:
                with open(os.path.join(ArchTable.ws, ArchTable.name + ".fac"), "rb") as f:
                    ArchTable.Geol.facies_domains=pickle.load(f)
                print("#### Facies loaded ####")
                ArchTable.facies_computed=1
            except:
                print("Facies results file not found")
        if props is None:
            #prop
            try:
                with open(os.path.join(ArchTable.ws, ArchTable.name + ".pro"), "rb") as f:
                    ArchTable.Geol.prop_values=pickle.load(f)
                print("#### Properties loaded ####")
                ArchTable.prop_computed=1
            except:
                print("Property results file not found")
            
    if surfs is not None:
        try:
            with open(surfs, "rb") as f:
                ArchTable.Geol.surfaces_by_piles=pickle.load(f)
                print("#### Surfaces loaded ####")
                ArchTable.surfaces_computed=1
        except:
            print("Surface results file not found")
    if surfs_bot is not None:
        try:
            with open(surfs_bot, "rb") as f:
                ArchTable.Geol.surfaces_bot_by_piles=pickle.load(f)      
                print("#### Bottom surfaces loaded ####")      
                ArchTable.surfaces_computed=1
        except:
            print("Surface bot results file not found")

    if units is not None:
        try:
            with open(units, "rb") as f:
                ArchTable.Geol.units_domains=pickle.load(f)  
                print("#### Units loaded ####")
                ArchTable.surfaces_computed=1
        except:
            print("Units results file not found")
    if facies is not None:
        try:
            with open(facies, "rb") as f:
                ArchTable.Geol.facies_domains=pickle.load(f)
                print("#### Facies loaded ####")
                ArchTable.facies_computed=1
        except:
            print("Facies results file not found")
    if props is not None:
        try:
            with open(props, "rb") as f:
                ArchTable.Geol.prop_values=pickle.load(f)  
                print("#### Properties loaded ####")
                ArchTable.prop_computed=1
        except:
            print("Property results file not found")

def load_bh_files(list_bhs, facies_data, units_data,
                  lbhs_bh_id_col="bh_ID", u_bh_id_col="bh_ID", fa_bh_id_col="bh_ID",
                  u_top_col="top",u_bot_col="bot",u_ID="Strat",
                  fa_top_col="top",fa_bot_col="bot",fa_ID="facies_ID",
                  bhx_col="bh_x", bhy_col='bh_y', bhz_col='bh_z', bh_depth_col='bh_depth',
                  dic_units_names={},
                  dic_facies_names={}, altitude=True, vb=1):

    """
    This function merges unit and facies
    databases in only one, 
    each row of the output dataframe corresponds
    to a layer information about units and facies
    
    Parameters
    ----------
    list_bhs : panda dataframe
        list of boreholes in ArchPy format
    facies_data : panda dataframe
        facies data file in ArchPy format
    units_data : panda dataframe
        unit data file in ArchPy format
    lbhs_bh_id_col : str, optional
        column name of borehole identifier in list_bhs data file, by default "bh_ID"
    u_bh_id_col : str, optional
        column name of borehole identifier in unit data file, by default "bh_ID"
    fa_bh_id_col : str, optional
        column name of borehole identifier in facies data file, by default "bh_ID"
    u_top_col : str, optional
        column name of top elevation info in unit data file, by default "top"
    u_bot_col : str, optional
        column name of bot elevation info in unit data file, by default "bot"
    u_ID : str, optional
        column name of unit identifier info in unit data file, by default "Strat"
    fa_top_col : str, optional  
        column name of top elevation info in facies data file, by default "top"
    fa_bot_col : str, optional  
        column name of bot elevation info in facies data file, by default "bot"
    fa_ID : str, optional
        column name of top elevation info in facies data file, by default "facies_ID"
    dic_units_names : dictionary, optional
        dictionary of units names, by default {}
    dic_facies_names : dictionary, optional
        dictionary of facies names, by default {}
    altitude : bool, optional
        if True, the elevation of the boreholes is taken into account, by default True
    vb : int, optional
        verbose, by default 1

    Returns
    -------
    panda dataframe
        dataframe with unit and facies information merged
    """

    def merge_dbs (fa_data, s_data):

        strat=[]
        for bh_id in list_bhs.index:
            
            if bh_id in fa_data.index and bh_id in s_data.index:
                fa_idata=fa_data.loc[[bh_id]].reset_index()
                s_idata=s_data.loc[[bh_id]].reset_index()
                
                #first top and bots
                ifa_top=fa_idata.top.values[0]
                ifa_bot=fa_idata.bot.values[0]
                is_top=s_idata.top.values[0]
                is_bot=s_idata.bot.values[0]
                
                #set top of log to top of bh
                if fa_idata.top.values[0] != list_bhs.loc[bh_id]["bh_z"]:
                    fa_idata.top.values[0]=list_bhs.loc[bh_id]["bh_z"]
                if s_idata.top.values[0] != list_bhs.loc[bh_id]["bh_z"]:
                     s_idata.top.values[0]=list_bhs.loc[bh_id]["bh_z"]
                     
                i_fa=0
                i_s=0
                ibot=None
                itop=None
                for i in range(fa_idata.shape[0]+s_idata.shape[0]):  # loop over layers in borehole

                    #top, bot and id from facies and unit data 
                    ifa_top=fa_idata.top.values[i_fa]
                    ifa_bot=fa_idata.bot.values[i_fa]
                    ifa_id=fa_idata.facies_ID.values[i_fa]

                    is_top=s_idata.top.values[i_s]
                    is_bot=s_idata.bot.values[i_s]
                    is_id=s_idata.Strat.values[i_s]
                    
                    if itop is None:
                        itop=is_top
                    else:
                        itop=ibot
                        
                    ibot=max(is_bot, ifa_bot) #take highest bot
                    strat.append((bh_id, is_id, ifa_id, itop, ibot)) #add interval
                    
                    if ifa_bot > is_bot:
                        i_fa += 1

                    elif ifa_bot < is_bot:
                        i_s += 1

                    else: #if both bot match
                        i_s += 1
                        i_fa += 1

                    if i_s == s_idata.shape[0]: #end of the borehole
                            break
                    if i_fa == fa_idata.shape[0]:
                            break

            elif bh_id in s_data.index: # if there is info about unit but not facies --> add none to facies
                
                s_idata=s_data.loc[[bh_id]].reset_index()
                
                for i in range(s_idata.shape[0]):
                    is_top=s_idata.top.values[i]
                    is_bot=s_idata.bot.values[i]
                    is_id=s_idata.Strat.values[i]
                    
                    strat.append((bh_id, is_id, None, is_top, is_bot))

            elif bh_id in fa_data.index:# if there is info about faces but not about unit --> add none to unit
                
                fa_idata=fa_data.loc[[bh_id]].reset_index()
                
                for i in range(fa_idata.shape[0]):
                    ifa_top=fa_idata.top.values[i]
                    ifa_bot=fa_idata.bot.values[i]
                    ifa_id=fa_idata.facies_ID.values[i]
                    
                    strat.append((bh_id, None, ifa_id, ifa_top, ifa_bot))
            

        all_strats=pd.DataFrame(strat, columns=["bh_ID","Strat_ID","Facies_ID","top","bot"]).set_index("bh_ID")
        return all_strats

    #fill gaps
    def add_gap(data, bh_id, top, bot):
        if vb:
            print("Gap encountered - creation of a gap interval")
        idata=data.loc[[bh_id]]
        lay=idata.iloc[0]
        lay.top=top
        lay.bot=bot
        lay.facies_ID=None
        lay.Strat=None
        data=data.append(pd.DataFrame(lay).T)
        return data

    #loading files
    #fa_data=pd.read_csv(facies_data)
    fa_data=facies_data.copy()
    fa_data.rename(columns={fa_bh_id_col:"bh_ID", fa_top_col:"top",fa_bot_col:"bot",fa_ID:"facies_ID"},inplace=True)
    if altitude:
        fa_data.sort_values(by=["bh_ID","top"], ascending=False, inplace=True)
    else:
        fa_data.sort_values(by=["bh_ID","top"], ascending=True, inplace=True)

    fa_data.set_index("bh_ID", inplace=True)
    
    #s_data=pd.read_csv(units_data)
    s_data=units_data.copy()
    s_data.rename(columns={u_bh_id_col:"bh_ID", u_top_col:"top",u_bot_col:"bot",u_ID:"Strat"},inplace=True)
    if altitude:
        s_data.sort_values(by=["bh_ID","top"], ascending=False, inplace=True)
    else:
        s_data.sort_values(by=["bh_ID","top"], ascending=True, inplace=True)

    s_data.set_index("bh_ID", inplace=True)
    
    #list_bhs=pd.read_csv(list_bhs)
    list_bhs=list_bhs.copy()
    list_bhs.rename(columns={lbhs_bh_id_col:"bh_ID", bhx_col:"bh_x", bhy_col:"bh_y", bhz_col:"bh_z", bh_depth_col:"bh_depth"}, inplace=True)
    list_bhs=list_bhs[["bh_ID", "bh_x", "bh_y", "bh_z", "bh_depth"]].copy() #keep only wanted columns
    list_bhs.set_index("bh_ID", inplace=True)

    #apply dictionnaries to changes identifier 
    fa_data.replace(dic_facies_names, inplace=True)
    s_data.replace(dic_units_names, inplace=True)
    
    #change depth into altitude
    if not altitude:
        for bh_id in list_bhs.index:
            top_0=list_bhs.loc[bh_id, "bh_z"] #borehole altitude
            if bh_id in fa_data.index:
                fa_data.loc[bh_id, "top"]=top_0 - fa_data.loc[bh_id, "top"]
                fa_data.loc[bh_id, "bot"]=top_0 - fa_data.loc[bh_id, "bot"]

            if bh_id in s_data.index:
                s_data.loc[bh_id, "top"]=top_0 - s_data.loc[bh_id, "top"]
                s_data.loc[bh_id, "bot"]=top_0 - s_data.loc[bh_id, "bot"]
                
    #fill gaps
    for bh_id in list_bhs.index:

        bh=list_bhs.loc[bh_id]

        if bh_id in fa_data.index:
            fa_idata=fa_data.loc[[bh_id]].reset_index()
        else:
            fa_idata=np.array(())

        if bh_id in s_data.index:
            s_idata=s_data.loc[[bh_id]].reset_index()
        else:
            s_idata =np.array(())

        top_0=bh["bh_z"]
        depth=bh["bh_depth"]

        #facies data
        for i in range(fa_idata.shape[0]):
            #top

            top=fa_idata.loc[i, "top"]
            if top_0 != top and i == 0:

                if vb:
                    print("Error, top altitude of first facies does not match borehole altitude")
                #fa_idata.loc[i,"top"]=top_0
                fa_data=add_gap(fa_data, bh_id, top_0, top)

            if i > 0 and bot != top:  # if there is a gap in the data
                fa_data=add_gap(fa_data, bh_id, bot, top)

            bot=fa_idata.loc[i,"bot"]

            if i == fa_idata.shape[0] - 1:  # last lay
                if (bot-1e-5 > top_0 - depth):  # if bot above maximum depth of borehole --> add a gap
                    fa_data=add_gap(fa_data, bh_id, bot, top_0 - depth)

        #units data
        for i in range(s_idata.shape[0]):

            #top
            top=s_idata.loc[i, "top"]

            if top_0 != top and i == 0:
                if vb:
                    print("Error, top altitude of first unit does not match borehole altitude")

                #s_idata.loc[i,"top"]=top_0
                s_data=add_gap(s_data, bh_id, top_0, top)

            if i > 0 and bot != top:  # if there is a gap in the data
                s_data=add_gap(s_data, bh_id, bot, top)

            bot=s_idata.loc[i,"bot"]

            if i == s_idata.shape[0] - 1:  # last lay
                if (bot-1e-5 > top_0 - depth):  # if bot above maximum depth of borehole --> add a gap
                    s_data=add_gap(s_data, bh_id, bot, top_0 - depth)
                    print(1)

    fa_data.index.rename("bh_ID", inplace=True)
    s_data.index.rename("bh_ID", inplace=True)

    #sort everything
    fa_data=fa_data.reset_index().sort_values(by=["bh_ID","top"],ascending=False).set_index("bh_ID")
    s_data=s_data.reset_index().sort_values(by=["bh_ID","top"],ascending=False).set_index("bh_ID")

    #merge
    final_db=merge_dbs(fa_data, s_data)

    return final_db, list_bhs


def extract_bhs(df, list_bhs, ArchTable, units_to_ignore=(), facies_to_ignore=(), extract_units=True, extract_facies=True, ignore2None=True, fill_gaps=True, vb=0):
     
    """
    Return a list of ArchTable boreholes from final database of load_bh_files

    Parameters
    ----------
    df : pandas dataframe
        final database of load_bh_files
    list_bhs : pandas dataframe
        list of boreholes dataframe with appropriate column name (generally use second output of :func:`load_bh_files`)
    ArchTable : :class:`base.Arch_table` object
        Arch_table object
    units_to_ignore : sequence of string, optional
        unit identifiers to ignore in the database df. The default is ().
    facies_to_ignore : sequence of string, optional
        facies identifiers to ignore in the database df. The default is ().
    extract_units : bool, optional
        extracting unit info. The default is True.
    extract_facies : bool, optional
        extracting facies info. The default is True.
    ignore2None : bool, optional
        Ignoring units and facies are considered as None (i.e. gaps). The default is True.
    fill_gaps : bool, optional
        if possible, gaps are filled. It means that if you have a gap in the data, 
        the unit/facies is considered as the one above. The default is True.
    vb : int, optional
        verbose. The default is 0.

    Returns
    -------
    list of :class:`base.borehole` objects 
        list of boreholes objects

    """
 
    #list_bhs=pd.read_csv(list_bhs)

    #reset index for getting boreholes objects
    df=df.reset_index()
    list_bhs=list_bhs.reset_index()
    
    #get boreholes
    i=0
    idepth=0
    bh_ID_prev=""
    unit_name_prev=""
    facies_name_prev= ""
    prev_ignore_flag=False
    l_bhs=[]
    log_strati=[]
    log_facies=[]

    altitude=True

    for idx in df.index:
    
        bh_ID=df.loc[idx,"bh_ID"]
        

        #store borehole
        if bh_ID != bh_ID_prev:  # new borehole --> save old one
            if log_strati or log_facies:  # if data about units or facies 
                
                # check if there is at least one info in each log
                if len([s[0] for s in log_strati if s[0] is not None]) == 0:
                    log_strati=None
                if len([s[0] for s in log_facies if s[0] is not None]) == 0:
                    log_facies=None
                
                if log_strati is not None or log_facies is not None: 
                   
                    # get borehole info
                    bh=list_bhs.loc[list_bhs["bh_ID"] == bh_ID_prev] 
                    x=bh["bh_x"].values[0]
                    y=bh["bh_y"].values[0]
                    z=bh["bh_z"].values[0]
                    depth=bh["bh_depth"].values[0]
                    if idepth != depth:
                        if vb:
                            print("end of data does not match end of borehole. \nBorehole depth adapted to the data")
                        depth=idepth
                    ID=bh["bh_ID"].values[0]
                    l_bhs.append(ArchPy.base.borehole(i, ID, x, y, z, depth, log_strati, log_facies))
                    i += 1

                    #reinitialize lith and facies names
                    unit_name_prev=""
                    facies_name_prev= ""
                    prev_ignore_flag=False

                    #reini idepth
                    idepth=0
                
            bh_ID_prev=bh_ID

            #reset logs
            log_strati=[]
            log_facies=[]

        if extract_units:
            unit_name=df.loc[idx,"Strat_ID"]
            #log strati
            if unit_name not in units_to_ignore:
                prev_ignore_flag=False
                if unit_name in [i.name for i in ArchTable.get_all_units()] or unit_name is None:# new bh that have strati info
                    #if unit_name not in [i[0].name for i in log_strati if i[0] is not None]: # if this encountered unit is not already in log
                    if unit_name_prev != unit_name:
                        if altitude:
                            z=df.loc[idx,"top"] 
                        else:
                            z=list_bhs.loc[list_bhs["bh_ID"] == bh_ID]["bh_z"].values[0] - df.loc[idx,"top"] 
                        z=np.round(z, 2)

                        if unit_name is not None:
                            log_strati.append((ArchTable.get_unit(unit_name),z))
                        else:
                            log_strati.append((None, z))
                        unit_name_prev=unit_name
                        
                    #idepth --> variable to ensure that the depth of the bh will be reached
                    if (list_bhs.loc[list_bhs["bh_ID"] == bh_ID]["bh_z"].values[0] - df.loc[idx,"bot"]) > idepth:
                        idepth=list_bhs.loc[list_bhs["bh_ID"] == bh_ID]["bh_z"].values[0] - df.loc[idx,"bot"]

                        
                else:
                    if vb:
                        print("{} unit name not found in ArchTable".format(unit_name))
                    pass
            else:  # unit to ignore
                if ignore2None:
                    if prev_ignore_flag:
                        pass
                    else:
                        if altitude:
                            z=df.loc[idx,"top"] 
                        else:
                            z=list_bhs.loc[list_bhs["bh_ID"] == bh_ID]["bh_z"].values[0] - df.loc[idx,"top"]
                        z=np.round(z, 2)

                        if log_strati and fill_gaps:
                            if log_strati[-1][0] is None:  # if latest entry is None
                                log_strati.pop(-1)  # remove latest None entry

                        log_strati.append((None, z))
                        prev_ignore_flag = True

                        if not fill_gaps:
                            unit_name_prev = unit_name

        if extract_facies:
            facies_name=df.loc[idx,"Facies_ID"]
       
            #log_facies
            if facies_name not in facies_to_ignore:
                prev_ignore_flag = False
                if facies_name in [i.name for i in ArchTable.get_all_facies()] or facies_name is None:# new bh that have strati info
                    if facies_name_prev != facies_name:
                        if altitude:
                            z=df.loc[idx,"top"] 
                        else:
                            z=list_bhs.loc[list_bhs["bh_ID"] == bh_ID]["bh_z"].values[0] - df.loc[idx,"top"] 
                        z=np.round(z, 2)
                        
                        #idepth --> variable to ensure that the depth of the bh will be reached
                        if list_bhs.loc[list_bhs["bh_ID"] == bh_ID]["bh_z"].values[0] - df.loc[idx,"bot"] > idepth:
                            idepth=list_bhs.loc[list_bhs["bh_ID"] == bh_ID]["bh_z"].values[0] - df.loc[idx,"bot"]
                            
                        if facies_name is not None:
                            log_facies.append((ArchTable.get_facies_obj(facies_name),z))
                        else:
                            log_facies.append((None, z))
                        facies_name_prev=facies_name
                else:
                    if vb:
                        print("{} facies name not found in ArchTable".format(facies_name))
            else:  # unit to ignore
                if ignore2None:
                    if prev_ignore_flag:
                        pass
                    else:
                        if altitude:
                            z=df.loc[idx,"top"] 
                        else:
                            z=list_bhs.loc[list_bhs["bh_ID"] == bh_ID]["bh_z"].values[0] - df.loc[idx,"top"]
                        z=np.round(z, 2)

                        if log_facies and fill_gaps:
                            if log_facies[-1][0] is None:  # if latest entry is None
                                log_facies.pop(-1)  # remove latest None entry

                        log_facies.append((None, z))
                        prev_ignore_flag = True

                        if not fill_gaps:
                            facies_name_prev = facies_to_ignore
                

    #store final borehole
    if log_strati or log_facies: #if data about units or facies (to do)
        
        #check if there is at least one info in each log
        if len([s[0] for s in log_strati if s[0] is not None]) == 0:
            log_strati=None
        if len([s[0] for s in log_facies if s[0] is not None]) == 0:
            log_facies=None
        
        # get borehole info
        bh=list_bhs.loc[list_bhs["bh_ID"] == bh_ID_prev] 
        x=bh["bh_x"].values[0]
        y=bh["bh_y"].values[0]
        z=bh["bh_z"].values[0]
        depth=bh["bh_depth"].values[0]
        ID=bh["bh_ID"].values[0]
        l_bhs.append(ArchPy.base.borehole(i, ID, x, y, z, depth, log_strati, log_facies))
        i += 1
    
    return l_bhs

### import ###
def import_d_facies(dic_project):
    
    """
    Import a dictionary of facies objects from the dictionary of
    the project that was imported from the yaml file

    Parameters
    ----------
    dic_project : dict
        Dictionary of the project imported from the yaml file

    Returns
    -------
    dict
        Dictionary of facies objects
    """
    d={}

    if "Facies" not in dic_project.keys():
        print("Facies field not defined in the project yaml file")
        return d
    
    dic_facies=dic_project["Facies"]
    for k, v in dic_facies.items():
        d[k]=ArchPy.base.Facies(v["ID"], k, v["color"])
        
    return d


def import_d_units(dic_project, all_facies, ws=None):

    """
    Import a dictionary of unit objects

    Parameters
    ----------
    dic_project : dict
        Dictionary of the project imported from the yaml file
    all_facies : dict
        Dictionary of all facies objects
    ws : str, optional
        Path to the working directory, by default None

    Returns
    -------
    dict
        Dictionary of units objects   
    """

    d={}
    dic_units=dic_project["Units"]
    for name, para in dic_units.items():

        #surface
        dic_s={}
        for k, v in para["surface"]["dic_surf"].items():
            if k == "covmodel":
                if v is not None:
                    dic_s[k]=ArchPy.inputs.imp_cm(v)
                else:
                    dic_s[k]=v
            else:
                dic_s[k]=v
        S=ArchPy.base.Surface(name=para["surface"]["name"],
                                contact= para["surface"]["contact"],
                                dic_surf=dic_s)    

        #dic_facies
        d_fa={}
        for k, v in para["dic_facies"].items():

            if k =="TI":
                if isinstance(v, str):
                    d_fa[k]=geone.img.readImageGslib(os.path.join(ws, v))
                else: #if not a path
                    d_fa[k]=v
            elif k == "G_cm" or k=="f_covmodel":
                if v is not None:
                    l_cms=[]
                    try:
                        for cm in v:
                            l_cms.append(ArchPy.inputs.imp_cm(cm))
                    except:
                        l_cms.append(ArchPy.inputs.imp_cm(v))

                    d_fa[k]=l_cms
                else:
                    d_fa[k]=v

            elif isinstance(v, str) and v.split(".")[-1] == "npy":
                d_fa[k]=np.load(os.path.join(ws, v))

            else:
                d_fa[k]=v
        unit=ArchPy.base.Unit(para["name"], order=para["order"], color=para["color"], ID=para["ID"], surface=S, dic_facies=d_fa)
        for fa in para["list_facies"]:
            unit.add_facies(all_facies[fa]) # add facies
        d[name]=unit
                
    return d

def import_d_prop(dic_project, d_fa):
    
    """
    Import a dictionary of property objects

    Parameters
    ----------
    dic_project : dict
        Dictionary of the project imported from the yaml file
    d_fa : dict
        Dictionary of facies objects (see :func:`import_d_facies`)
    """

    assert "properties" in dic_project.keys(), '"properties" field not defined in the project yaml file'
    
    d={}
    for name, para in dic_project["properties"].items():
        
            
        l_cms=[]
        for cm in para["covmodels"]:
            if cm is not None:
                l_cms.append(imp_cm(cm))
            else:
                l_cms.append(cm)

        l_facies=[d_fa[i] for i in para["facies"]] #get list of facies object
        
        if "x" not in para.keys():
            para["x"]=None
        if "v" not in para.keys():
            para["v"]=None
        Prop=ArchPy.base.Prop(para["name"], facies=l_facies, covmodels=l_cms,
                                means=para["means"], int_method=para["int_method"], def_mean=para["def_mean"], 
                                vmin=para["vmin"], vmax=para["vmax"], x=para["x"], v=para["v"])
    
        d[name]=Prop
    return d
                

def import_d_piles(dic_project, d_units):
    
    """
    Import a dictionary of pile objects from the given dictionary project

    Parameters
    ----------
    dic_project : dict
        Dictionary of the project imported from the yaml file
    d_units : dict
        Dictionary of units objects (see :func:`import_d_units`)
    """

    d={}
    
    dic_piles=dic_project["Piles"]
    for k, v in dic_piles.items():
        pile=ArchPy.base.Pile(v["name"], verbose=v["verbose"], seed=v["seed"])
        for unit in v["list_units"]:
            pile.add_unit([d_units[unit]]) # to finish
        d[k]=pile
    return d


def import_project(project_name, ws, import_bhs=True, import_results=True, import_grid=True):
    
    """
    Load an ArchPy project (Arch_table) 
    with the specified name (project_name) 
    and in the working_directory (ws) 
    
    Note
    ----
    All the configurations files must be in the working directory

    Parameters
    ----------
    project_name : str
        Name of the project --> name of the yaml file
    ws : str
        Path to the working directory where all files are stored
    import_bh : bool, optional
        Flag to indicate to import borehole files, by default True
    import_results : bool, optional
        Flag to indicate to import result files, by default True
    import_grid : bool, optional
        Flag to indicate to import grid, by default True

    Returns
    -------
    :class:`ArchPy.base.Arch_table` object
    """
    
    print("### IMPORTING PROJECT {} IN {} DIRECTORY ### \n".format(project_name, ws))
    
    with open(os.path.join(ws, project_name+".yaml"), "r") as f:
        dic_project=yaml.load(f, Loader=yaml.Loader)  # bug fix latest yaml version
    
    # load archpy objects
    d_fa=import_d_facies(dic_project)  # facies
    d_un=import_d_units(dic_project, all_facies=d_fa, ws=ws)  # unit
    d_pro=import_d_prop(dic_project, d_fa=d_fa)  # prop
    d_piles=import_d_piles(dic_project, d_units=d_un)  # piles
    
    
    # add subpiles to units
    for unit in d_un.values():
        if "SubPile" in unit.dic_facies.keys():
            if unit.dic_facies["SubPile"] is not None:
                unit.set_SubPile(d_piles[unit.dic_facies["SubPile"]])
    
    ArchTable=ArchPy.base.Arch_table(dic_project["name"], working_directory=ws, seed = min(int(dic_project["seed"] / 1e6), 1),
                                      verbose=dic_project["verbose"], ncpu=dic_project["ncpu"])
    
    ArchTable.set_Pile_master(d_piles[dic_project["Pile_master"]])
    
    # add properties
    for prop in d_pro.values():
        ArchTable.add_prop(prop)

    if import_grid:
        #grid
        assert "grid" in dic_project.keys(), "Grid was not defined for project {}".format(project_name)
        di=dic_project["grid"]["dimensions"]
        spa=dic_project["grid"]["spacing"]
        ori=dic_project["grid"]["origin"]
        
        #top
        top=dic_project["grid"]["top"]
        if isinstance(top, str): #path
            if top.split(".")[-1] == "top":
                path=os.path.join(ws, top+".npy")
                if os.path.exists(path):
                    top=np.load(path)
                else:
                    print("Top array has not been found")

        #bot
        bot=dic_project["grid"]["bot"]
        if isinstance(bot, str): #path
            if bot.split(".")[-1] == "bot":
                path=os.path.join(ws, bot+".npy")
                if os.path.exists(path):
                    bot=np.load(path)
                else:
                    print("bot array has not been found")
        
        # add polygon
        # mask
        mask = dic_project["grid"]["mask"]
        if mask.split(".")[-1] == "msk":
            path=os.path.join(ws, mask+".npy")
            if os.path.exists(path):
                mask=np.load(path)

        di=(int(di[0]), int(di[1]), int(di[2]))
        ArchTable.add_grid(di, spa, ori, top=top, bot=bot, mask=mask)
    
    if import_bhs:
        # boreholes
        if len(dic_project["boreholes"]) > 0:  # if there is borehole info
            l_bh_path=os.path.join(ws, dic_project["boreholes"]["list_bhs"])
            fd_path=os.path.join(ws, dic_project["boreholes"]["facies_data"])
            un_path=os.path.join(ws, dic_project["boreholes"]["units_data"])
            l_bhs=pd.read_csv(l_bh_path)
            fa_data=pd.read_csv(fd_path)
            un_data=pd.read_csv(un_path)
            db, l_bhs=ArchPy.inputs.load_bh_files(l_bhs, fa_data, un_data, altitude=True)
            
            boreholes=extract_bhs(db, l_bhs, ArchTable)
        else:
            print("No borehole data found")
        
        ArchTable.add_bh(boreholes)
    
    if "geol_map" in dic_project.keys():
        path=os.path.join(ws, dic_project["geol_map"]+".npy")
        if os.path.exists(path):
            geol_map=np.load(path)
            ArchTable.add_geological_map(geol_map)

    if import_results:
        print("\n\n ##LOADING RESULTS## \n\n")
        # load results 
        if "Results" in dic_project.keys():
            if dic_project["Results"] is not None:
                d_res=dic_project["Results"]
                if "surfaces" in d_res.keys():
                    surfs=os.path.join(ws, d_res["surfaces"])
                else:
                    surfs=None
                if "surfaces_bot" in d_res.keys():
                    surfs_bot=os.path.join(ws, d_res["surfaces_bot"])
                else:
                    surfs_bot=None
                # if "units" in d_res.keys():
                #     units=os.path.join(ws, d_res["units"])
                # else:
                #     units=None
                # if "facies" in d_res.keys():
                #     facies=os.path.join(ws, d_res["facies"])
                # else:
                #     facies=None
                # if "properties" in d_res.keys():
                #     props=os.path.join(ws, d_res["properties"])
                # else:
                #     props=None

                load_results(ArchTable, surfs, surfs_bot)
            else:
                print("No results files provided in the yaml file \n")
                print("Trying to find results files in working_directory")
                load_results(ArchTable)
            
            
    print("\n\n ### SUCCESSFUL IMPORT ### \n")
    return ArchTable


#### Write files ####
def write_bh_files(ArchTable):
    
    """
    Write input files from an existing ArchTable project
    """

    # dictionnaries
    l_bh={}
    stratis={}
    facies={}
    
    # empty lists
    l_id=[]
    l_name=[]
    l_bhx=[]
    l_bhy=[]
    l_bhz=[]
    l_depth=[]
    
    strati_bhid=[]
    strati_id=[]
    strati_top=[]
    strati_bot=[]
    
    facies_bhid=[]
    facies_id=[]
    facies_top=[]
    facies_bot=[]
    
    for bh in ArchTable.list_bhs:
        l_id.append(bh.ID)
        l_name.append(bh.name)
        l_bhx.append(bh.x)
        l_bhy.append(bh.y)
        l_bhz.append(bh.z)
        l_depth.append(bh.depth)
        
        #extract log_strati
        if bh.log_strati is not None:
            for i, s in enumerate(bh.log_strati):
                unit=s[0]
                top=s[1]
                if i < len(bh.log_strati) - 1:
                    bot=bh.log_strati[i+1][1]
                else:
                    bot=bh.z - bh.depth
                
                strati_bhid.append(bh.ID)
                if unit is not None:
                    strati_id.append(unit.name)
                else:
                    strati_id.append(None)
                strati_top.append(top)
                strati_bot.append(bot)
                
        if bh.log_facies is not None:
            for i, s in enumerate(bh.log_facies):
                fa=s[0]
                top=s[1]
                if i < len(bh.log_facies) - 1:
                    bot=bh.log_facies[i+1][1]
                else:
                    bot=bh.z - bh.depth
                
                facies_bhid.append(bh.ID)
                if fa is not None:
                    facies_id.append(fa.name)
                else:
                    facies_id.append(None)
                facies_top.append(top)
                facies_bot.append(bot)  
    
    #dic list boreholes
    l_bh["bh_ID"]=l_id
    l_bh["bh_x"]=l_bhx
    l_bh["bh_y"]=l_bhy
    l_bh["bh_z"]=l_bhz
    l_bh["bh_depth"]=l_depth
    
    stratis["bh_ID"]=strati_bhid
    stratis["Strat"]=strati_id
    stratis["top"]=strati_top
    stratis["bot"]=strati_bot

    facies["bh_ID"]=facies_bhid
    facies["facies_ID"]=facies_id
    facies["top"]=facies_top
    facies["bot"]=facies_bot    
    
    pd.DataFrame(l_bh).set_index("bh_ID").to_csv(os.path.join(ArchTable.ws, ArchTable.name+".lbh")) #list boreholes
    pd.DataFrame(stratis).set_index("bh_ID").to_csv(os.path.join(ArchTable.ws, ArchTable.name+".ud")) #units data
    pd.DataFrame(facies).set_index("bh_ID").to_csv(os.path.join(ArchTable.ws, ArchTable.name+".fd")) #facies data
    
    return ArchTable.name+".lbh", ArchTable.name+".ud", ArchTable.name+".fd"

def save_results(ArchTable):

    """
    Save ArchTable results
    """

    d={}
    if ArchTable.surfaces_computed:
        #surfaces
        fn=ArchTable.name + ".sf"
        d["surfaces"]=fn
        with open(os.path.join(ArchTable.ws, fn), "wb") as f:
            pickle.dump(ArchTable.Geol.surfaces_by_piles, f)

        #surfaces bot   
        fn=ArchTable.name + ".sfb"
        d["surfaces_bot"]=fn
        with open(os.path.join(ArchTable.ws, fn), "wb") as f:
            pickle.dump(ArchTable.Geol.surfaces_bot_by_piles, f)
    
    if ~ArchTable.write_results:
        #units
        fn=ArchTable.name + ".unt"
        d["units"]=fn
        with open(os.path.join(ArchTable.ws, fn), "wb") as f:
            pickle.dump(ArchTable.Geol.units_domains, f)

        if ArchTable.facies_computed:
            #facies
            fn=ArchTable.name + ".fac"
            d["facies"]=fn
            with open(os.path.join(ArchTable.ws, fn), "wb") as f:
                pickle.dump(ArchTable.Geol.facies_domains, f)

        if ArchTable.prop_computed:
            #prop
            fn=ArchTable.name + ".pro"
            d["properties"]=fn 
            with open(os.path.join(ArchTable.ws, fn), "wb") as f:
                pickle.dump(ArchTable.Geol.prop_values, f)

    return d

# Create data in a format acceptable for yaml#
def Cm2YamlCm (cm):

    """
    Convert covmodel numbers to float format for yaml compatibility

    :meta private:
    """

    cm_tr=[]
    for el in cm.elem:

        #new d
        new_d={}
        for k, v in el[1].items():
            v=np.array(v)
            if len(v.shape) > 0: 
                n_l=[]
                for iv in v:
                    n_l.append(float(iv))
                new_d[k]=n_l
            else:
                new_d[k]=float(v)     
        cm_tr.append((el[0],new_d))
    
    cm.elem=cm_tr
    if isinstance(cm, geone.covModel.CovModel2D):
        cm.alpha=float(cm.alpha)
    if isinstance(cm, geone.covModel.CovModel3D):
        cm.alpha=float(cm.alpha)
        cm.beta=float(cm.beta)
        cm.gamma=float(cm.gamma)
    return cm


#create dic funs
#l_num_types= (int, float) #all number types
#np_num_types=(np.float16, np.float32, np.float64, np.int0, np.int8, np.int16, np.int32, np.int64)

def create_d_facies(self):
    l={i.name: {"ID": i.ID, "color": i.c} for i in self.get_all_facies()}
    
    return l

def create_d_f_covmodels(unit):

    l=[{"alpha": i.alpha, "beta": i.beta, "gamma": i.gamma, "elem": Cm2YamlCm(i).elem} for i in unit.list_f_covmodel]
    
    return l

def create_d_surface(surface, ws):
    
    """
    create a yaml dic for a surface object

    :meta private:
    """
    d={}
    d["name"]=surface.name
    d["contact"]=surface.contact
    
    d_dic_s={}
    for k, v in surface.dic_surf.items():
        
        if isinstance(v, str) or isinstance(v, (int, float)):
            d_dic_s[k]=v
        elif isinstance(v, gcm.CovModel2D):
            cm=Cm2YamlCm(v)
            d_dic_s["covmodel"]={"alpha": float(cm.alpha), "elem": cm.elem}
        
        elif hasattr(v, "__iter__") and isinstance(v[0], (int, float)): #sequence of numbers
            new_v=[]
            for iv in v:
                new_v.append(float(iv))  # change data format cause yaml doesn't support np.float so to be sure...
            d_dic_s[k]=new_v
        
        elif isinstance(v, np.ndarray):  # numpy array --> store it in binary file
            f_name=surface.name + "_dic_surf_" + str(k) 
            np.save(os.path.join(ws, f_name),v)
            d_dic_s[k]=f_name
            
        elif isinstance(v, geone.img.Img):  # geone image --> just save file in ws
            f_name=surface.name + "_dic_surf_" + str(k)
            geone.img.writeImageGslib(v, os.path.join(ws, f_name))  # write file
            d_dic_s[k]=f_name
            
        else:  # warning some arguments cannot work with the export
            d_dic_s[k]=v
            
    d["dic_surf"]=d_dic_s
    return d

def create_d_unit(unit, ws):
    
    d={}
    d["name"]=unit.name
    d["order"]=unit.order
    
    if isinstance(unit.c, str):
        c=unit.c
        d["color"]=c
    elif isinstance(unit.c, tuple) or isinstance(unit.c, np.array) or isinstance(unit.color, list):
        new_c=[]
        for ic in unit.color:
            new_c.append(float(ic)) #change data format cause yaml doesn't support np.float
        d["color"]=new_c
    
    d["ID"]=unit.ID
    d["surface"]=create_d_surface(unit.surface, ws)
    d["list_facies"]=[i.name for i in unit.list_facies]

    #dic facies
    d_dic_f={}
    for k, v in unit.dic_facies.items():
        if isinstance(v, (str, int)):
            d_dic_f[k]=v
        if isinstance(v, float): #only a string, float or int
            d_dic_f[k]=float(v)
        elif isinstance(v, gcm.CovModel3D): #f covmodel
            cm=Cm2YamlCm(v)
            d_dic_f[k]={"alpha": cm.alpha, "beta": cm.beta, "gamma": cm.gamma, "elem": cm.elem}
        elif isinstance(v, (list, tuple)) and isinstance(v[0],gcm.CovModel3D): #list of fcovmodels
            l=[]
            for iv in v:
                cm=Cm2YamlCm(iv)
                l.append({"alpha": cm.alpha, "beta": cm.beta, "gamma": cm.gamma, "elem": cm.elem})
            d_dic_f[k]=l
        elif isinstance(v, geone.img.Img): #geone image    
            f_name=unit.name + "_dic_facies_" + str(k) + ".gslib"
            geone.img.writeImageGslib(v, os.path.join(ws, f_name)) #write file
            d_dic_f[k]=f_name
        
        elif isinstance(v, np.ndarray): #numpy array --> store it in binary file
            f_name=unit.name + "_dic_facies_" + str(k)
            np.save(os.path.join(ws, f_name),v) #create file
            d_dic_f[k]=f_name+".npy"

        elif isinstance(v, (tuple, list)) and isinstance(v[0], (int, float)): #sequence like of numbers
            l=[]
            for iv in v:
                l.append(float(iv))
            d_dic_f[k]=l
        
        elif k == "SubPile":
            if v is not None:
                d_dic_f[k]=v.name
            else:
                 d_dic_f[k]=v
        else:
            d_dic_f[k]=v
            
    d["dic_facies"]=d_dic_f
    
    return d

def create_d_pile(pile):


    d={}
    d["name"]=pile.name
    d["list_units"]=[i.name for i in pile.list_units]
    d["verbose"]=pile.verbose
    d["seed"]=pile.seed

    return d

def create_d_prop(prop):

    def fun(seq):
        l=[]
        for i in seq:
            if hasattr(i, "__iter__"):
                l.append([float(o) for o in i])
            else:
                l.append(float(i))
        return l

    def fun2(seq):
        l=[]
        for i in seq:
            if i is not None:
                l.append(float(i))
            else:
                l.append(i)
        return l
    d={}
    d["name"]=prop.name
    d["facies"]=[i.name for i in prop.facies]

    l=[]
    for i in prop.covmodels:
        if i is not None:
            l.append({"alpha": i.alpha, "beta": i.beta, "gamma": i.gamma, "elem": Cm2YamlCm(i).elem})
        else:
            l.append(None)
    d["covmodels"]=l

    d["means"]=fun2(prop.means)
    d["int_method"]=prop.int
    d["def_mean"]=prop.def_mean
    d["vmin"]=prop.vmin
    d["vmax"]=prop.vmax
    if prop.x is not None:
        d["x"]=fun(prop.x)
    if prop.v is not None:
        d["v"]=fun(prop.v)


    return d

def create_d_units(ArchTable):

    d={}
    for unit in ArchTable.get_all_units():
        d[unit.name]=ArchPy.inputs.create_d_unit(unit, ArchTable.ws)

    return d


def create_d_piles(ArchTable):

    d={}
    for pile in ArchTable.get_piles():
        d[pile.name]=create_d_pile(pile)

    return d


def create_d_properties(ArchTable):

    d={}
    for prop in ArchTable.list_props:
        d[prop.name]=create_d_prop(prop)

    return d

def save_project(ArchTable, results=True):
    
    """
    Save an Arch_table object in a yaml file and other supplementary files 
    (images, numpy arrays, gslib files) in a folder.

    Parameters
    ----------
    ArchTable : :class:`base.Arch_table`
        Arch_table object to save
    results : bool, optional
        flag to save results or not
    """

    d={}
    d["name"]=ArchTable.name
    d["ws"]=ArchTable.ws
    if ArchTable.ws not in os.listdir():
        os.makedirs(ArchTable.ws)
    d["seed"] = ArchTable.seed
    d["verbose"]=ArchTable.verbose
    d["Pile_master"]=ArchTable.get_pile_master().name
    d["ncpu"]=ArchTable.ncpu
    d["surfaces_computed"]=ArchTable.surfaces_computed
    d["facies_computed"]=ArchTable.facies_computed
    d["prop_computed"]=ArchTable.prop_computed

    ##grid##
    gr={}
    gr["dimensions"]=(int(ArchTable.get_nx()), int(ArchTable.get_ny()), int(ArchTable.get_nz()))
    gr["spacing"]=(float(ArchTable.get_sx()), float(ArchTable.get_sy()), float(ArchTable.get_sz()))
    gr["origin"]=(float(ArchTable.get_ox()), float(ArchTable.get_oy()), float(ArchTable.get_oz()))
    
    
    #top -> save top with numpy
    file=ArchTable.name+".top"
    np.save(os.path.join(ArchTable.ws, file), ArchTable.top)
    gr["top"]=file
    
    #bot -> save bot with numpy
    file=ArchTable.name+".bot"
    np.save(os.path.join(ArchTable.ws, file), ArchTable.bot)
    gr["bot"]=file
    
    #mask --> save mask with numpy
    file=ArchTable.name+".msk"
    np.save(os.path.join(ArchTable.ws, file), ArchTable.mask)
    gr["mask"]=file
    
#    file=ArchTable.name+".poly"
#    np.save(os.path.join(ArchTable.ws, file), ArchTable.mask.any(0))
#    gr["polygon"]=file

    #save grid
    d["grid"]=gr
    
    #geol map
    if ArchTable.geol_map is not None:
        file=ArchTable.name+".gmap"
        np.save(os.path.join(ArchTable.ws, file), ArchTable.geol_map)
        d["geol_map"]=file

    #save boreholes
    d_bh={}
    l_bh, units_data, facies_data=write_bh_files(ArchTable)
    d_bh["list_bhs"]=l_bh
    d_bh["units_data"]=units_data
    d_bh["facies_data"]=facies_data
    
    d["boreholes"]=d_bh
    
    #piles
    d["Piles"]=create_d_piles(ArchTable)

    #units
    d["Units"]=create_d_units(ArchTable)

    #facies
    d["Facies"]=create_d_facies(ArchTable)

    #properties
    d["properties"]=create_d_properties(ArchTable)

    #save results
    if results:
        d_res=save_results(ArchTable)
        if len(d_res) > 0:
            d["Results"]=d_res
        else:
            d["Results"]=None
    else:
        d["Results"]=None

    #create file
    with open(os.path.join(ArchTable.ws, ArchTable.name + ".yaml"), 'w') as file:
        documents=yaml.dump(d, file)

    print("Project saved successfully")

    return True 


def bhs_analysis(db,  Strat_ID="Strat_ID", Facies_ID="Facies_ID", top_col="top", bot_col="bot", ax=None):
    
    """
    Function to plot the facies proportions in boreholes and return the values in a dataframe

    Parameters
    ----------
    db : pandas dataframe
        dataframe containing the boreholes
    Strat_ID : str, optional
        name of the column containing the stratigraphic ID
    Facies_ID : str, optional
        name of the column containing the facies ID
    top_col : str, optional
        name of the column containing the top of the unit
    bot_col : str, optional
        name of the column containing the bottom of the unit
    ax : matplotlib axis, optional
        axis to plot the results

    Returns
    -------
    pandas dataframe   
        dataframe containing the facies proportions in each stratigraphic unit
    """

    t=db.copy()
    ##Facies
    t["thickness"]=t[top_col] - t[bot_col]
    
    
    print("### Units proportion in boreholes")
    print(t.groupby(["Strat_ID"])["thickness"].sum())
    
    
    if ax is None:
        fig, ax=plt.subplots(figsize=(15, 10))
    
    df=pd.DataFrame(t.groupby([Strat_ID, Facies_ID])["thickness"].sum()) #group thickness
    for s_id in df.index.get_level_values("Strat_ID").unique():
        df.loc[s_id]=df.loc[s_id].values/df.groupby("Strat_ID").sum().loc[s_id].values[0] #get proportions
    df.unstack()["thickness"].plot(kind="bar", stacked=True, legend=False, ax=ax)
    ax.set(ylabel="Facies proportion")
    ax.set_ylim(0, 1.5)
    ax.legend(loc='upper center', bbox_to_anchor=(0.5, 1),
              ncol=3, fancybox=True, shadow=True)
    plt.tight_layout()
    
    return df
    



"""
This module contains functions used for the validation of ArchPy models
For now it the validation mainly relies on cross-validation through the k-fold method
"""

import numpy as np
import matplotlib
from matplotlib import colors
import matplotlib.pyplot as plt
import pyvista as pv
import copy
import time
import sys
import pandas as pd

#geone
import geone
import geone.covModel as gcm
import geone.grf as grf
from geone import img
import geone.imgplot as imgplt
import geone.imgplot3d as imgplt3
import geone.deesseinterface as dsi
import geone.geosclassicinterface as gci

#ArchPy modules
import ArchPy
import ArchPy.base
from ArchPy.base import Geol

# functions

def brier_func(p, i):
    
    """
    Compute the Brier score given a vector of probabilities

    Parameters
    ----------

    p : sequence if float
        vector of probabilities. Values must be between 0 and 1
    i : int
        index of the true answer in vector

    Returns
    -------
    float
        Brier score
        
    """
    
    s=0
    M=len(p)
    for j in range(M):
        
        if i == j:
            s+= (1 - p[j])**2
        else:
            s+= (-p[j])**2
    
    return -s


def img_to_3D_colors(arr, dic):

    """
    Replace value in an array using a dictionnary
    linking actual values to new values

    Parameters
    ----------
    arr : ndarray
        array of values to replace
    dic : dict
        dictionnary linking actual values to new values

    Returns
    -------
    ndarray
        array of new values
    """

    arr_old=arr.flatten().copy()
    arr_new=np.ones([arr_old.shape[0], 3])*np.nan
    for i, x in enumerate(arr_old):
        if x==x:
            arr_new[i]=dic[x]
        else:
            #arr_new[i]=(np.nan, np.nan, np.nan)
            arr_new[i]=(1, 1, 1)
    return arr_new.reshape(*arr.shape, 3)


def bh2array(ArchTable, bh, typ="units"):

    """ 
    Function to convert an ArchPy borehole to an array of IDs, either units or facies
    IDs are given by the facies/units ID

    Parameters
    ----------
    ArchTable : :class:`ArchPy.base.Arch_table`
        ArchPy table
    bh : :class:`ArchPy.base.borehole`
        ArchPy borehole
    typ : str
        type of data to extract, either "units" or "facies"

    Returns
    -------
    ndarray
        array of IDs
    """
    
    step=ArchTable.get_sz()  # vertical resolution
    nz=ArchTable.get_nz()  # nz
    zg=ArchTable.get_zgc()  # z cell centers

    
    arr_un=np.ones([nz])*np.nan  # initialize units/facies array
    arr_fa=np.ones([nz])*np.nan
    
    if bh.log_facies is not None: 
        fa_flag=True
    else:
        fa_flag=False
    if bh.log_strati is not None:
        un_flag=True
    else:
        un_flag=False
    
    for iz in range(nz):
        z=zg[iz]
        extract = bh.extract(z, vb=0)
        if extract is None:
            pass
        elif fa_flag and un_flag:
            if extract[0] is not None:
                arr_un[iz] = extract[0].ID
            if extract[1] is not None:
                arr_fa[iz] = extract[1].ID
        elif fa_flag:
            if extract is not None:
                arr_fa[iz] = extract.ID
        elif un_flag:
            if extract is not None:
                arr_un[iz] = extract.ID
        else:
            raise ValueError("No facies or unit data")
    
    if typ=="units":
        return arr_un
    elif typ == "facies": 
        return arr_fa
    else: 
        return (arr_un, arr_fa)
    
    
def test_bh_1fold(ArchTable, bhs_real, bhs_test, weighting_method = "same_weight", dic_weights = None,
                  plot = True, brier = True, proba_correct = False, aspect="auto"):
    
    """
    Perform a test on a single fold of a cross-validation

    Parameters
    ----------
    ArchTable : :class:`ArchPy.base.Arch_table`
        ArchPy table to test
    bhs_real : sequence of :class:`ArchPy.base.borehole`
        "fake" boreholes sampled in the realizations of the ArchTable
    bhs_test : sequence of :class:`ArchPy.base.borehole`
        "real" boreholes to test
    weighting_method : str
        method to weight the realizations, either "same_weight" or "weights"
    dic_weights : dict
        dictionnary linking realization index to weight
    plot : bool
        plot or not
    brier : bool
        compute Brier score or not
    proba_correct : bool
        compute probability of correct classification or not
    aspect : str
        aspect ratio of the plot

    Returns
    -------
    dict
        dictionnary of results
    """
    
    reals = bhs_real
    scores_un = []
    scores_fa = []
    scores_correct_un = []
    scores_correct_fa = []

    for i in range(len(bhs_test)):
        bh_test = bhs_test[i]
        bh_real = reals[:, :, i]

        ## change units in realization logs to match unit in reference log
        for bh in bh_real:
            bh = bh[0]  # nested list
            for i in range(len(bh.log_strati)):
                s = bh.log_strati[i][0]
                something=0
                if s not in bh_test.get_list_stratis(): # if unit in real log is not in ref log
                    for unit in bh_test.get_list_stratis():
                        if unit is not None:  
                            if s in unit.get_baby_units(True, 0):  # check if there is a baby unit of any unit in ref log
                                bh.log_strati[i] = (unit, bh.log_strati[i][1])  # if so change baby to mummy unit

        ## list unit ids
        l = []
        def fun(pile, l):
            for un in pile.list_units:
                if un.f_method == "SubPile":
                    fun(un.SubPile, l)
                else:
                    l.append(un)
        fun(ArchTable.get_pile_master(), l)
        list_units_real = l  # list of all units that are simulated by ArchPy (highest level of hierarchy)

        list_ids = []  # list of possible ids for this reference log
        for i in range(len(list_units_real)):
            s = list_units_real[i]
            something=0
            if s not in bh_test.get_list_stratis(): # if unit in real log is not in ref log
                for unit in bh_test.get_list_stratis(): 
                    if unit is not None: 
                        if s in unit.get_baby_units(True, 0):  # check if there is a baby unit of any unit in ref log
                            if unit.ID not in list_ids:
                                list_ids.append(unit.ID)
                            something=1
                            break  

            if something == 0:  # if s (unit in real bh) is not at all in ref log then add it to list of possible outcomes
                if s.ID not in list_ids:
                        list_ids.append(s.ID)

        ## list facies ids
        list_facies_ids = [i.ID for i in ArchTable.get_all_facies()]

        # transform bh to array
        lu = []
        lfa = []
        for iu in range(ArchTable.nreal_units): 
            ifa = 0
            bh = bh_real[iu, ifa]
            g = bh2array(ArchTable, bh, typ="both")
            lu.append(g[0])
            lfa.append(g[1])

            for ifa in range(1, ArchTable.nreal_fa):
                bh = bh_real[iu, ifa]
                g = bh2array(ArchTable, bh, typ="facies")
                lfa.append(g)

        res_un = np.array(lu)
        res_fa = np.array(lfa)

        # personal WARNING when check borehole for X-valid - attention to hierarchical units !!!

        # ref
        ref_arr = bh2array(ArchTable, bh_test, typ="both") 
        ref_arr_un = ref_arr[0]
        ref_arr_fa = ref_arr[1]

        if plot:
            ## plot unit

            # dictionary for ID to rgb colors
            d = {}
            for unit in ArchTable.list_all_units:
                d[unit.ID] = matplotlib.colors.to_rgb(unit.c)


            # create rgb array for plotting
            res_plot = res_un.T.copy()
            res_plot = img_to_3D_colors(res_plot, d)
            ref_arr_plot = ref_arr_un.reshape(1, -1).T.copy()
            ref_arr_plot = img_to_3D_colors(ref_arr_plot, d)

            # 1st plot
            fig,ax = plt.subplots(1,2, figsize=(.5*ArchTable.nreal_units, 5), sharey=True)
            ax[0].imshow(res_plot, extent=[0, ArchTable.nreal_units, ArchTable.zg[0], ArchTable.zg[-1]], origin="lower", interpolation="nearest", aspect=aspect)
            ax[0].set_title("real")
            ax[0].set_xlabel("Realizations")
            ax[0].set_ylabel("Z [m]")

            ax[1].imshow(ref_arr_plot, extent=[0, ArchTable.nreal_units, ArchTable.zg[0], ArchTable.zg[-1]],
                       origin="lower", interpolation="nearest", aspect=aspect)
            ax[1].set_title("ref")
            plt.subplots_adjust(wspace=1/ArchTable.nreal_units)
            plt.show()

            if ArchTable.nreal_fa > 0:
                # facies
                dfa_c = {}
                for fa in ArchTable.get_all_facies():
                    dfa_c[fa.ID] = matplotlib.colors.to_rgb(fa.c)

                 # create rgb array for plotting
                res_plot = res_fa.T.copy()
                res_plot = img_to_3D_colors(res_plot, dfa_c)
                ref_arr_plot = ref_arr_fa.reshape(1, -1).T.copy()
                ref_arr_plot = img_to_3D_colors(ref_arr_plot, dfa_c)

                # 1st facies plot
                fig = plt.figure(figsize=(0.7*(ArchTable.nreal_fa+ArchTable.nreal_units), 1.5))
                ax1 = plt.subplot(1,2, 1)
                ax1.imshow(res_plot, extent=[0, (ArchTable.nreal_fa*ArchTable.nreal_units), ArchTable.zg[0], ArchTable.zg[-1]],
                           origin="lower", interpolation="nearest")
                plt.title("real")
                plt.xlabel("Realizations")
                plt.ylabel("Z [m]")

                ax2 = plt.subplot(1,2, 2, sharey = ax1)
                ax2.imshow(ref_arr_plot, extent=[0, ArchTable.nreal_units, ArchTable.zg[0], ArchTable.zg[-1]],
                           origin="lower", interpolation="nearest")
                plt.title("ref")
                plt.show()


        if proba_correct:
            ## compute_proba_correct

            # units
            model_res = []
            for ireal in range(res_un.shape[0]):
                imodel = res_un[ireal]
                mask = (ref_arr_un == ref_arr_un)
                n_correct = np.sum(ref_arr_un[mask] == imodel[mask]) / np.sum(mask)
                model_res.append(n_correct)

            scores_correct_un.append(np.mean(model_res))

            # facies
            model_res = []
            for ireal in range(res_fa.shape[0]):
                imodel = res_fa[ireal]
                mask = (ref_arr_fa == ref_arr_fa)
                n_correct = np.sum(ref_arr_fa[mask] == imodel[mask]) / np.sum(mask)
                model_res.append(n_correct)

            scores_correct_fa.append(np.mean(model_res))

        if brier:
            ## compute brier scores

            # probability vectors
            p_real_un = np.array([np.count_nonzero(res_un == i, 0)/ArchTable.nreal_units for i in list_ids]).T
            p_real_fa = np.array([np.count_nonzero(res_fa == i, 0)/(ArchTable.nreal_units*ArchTable.nreal_fa) for i in list_facies_ids]).T

            # units
            d={}
            for i,k in enumerate(list_ids):
                d[k] = i

            l = []
            ref = []
            for i in range(ref_arr_un.shape[0]):
                if ~np.isnan(ref_arr_un[i]):  # ignore nan in test bhs
                    ref.append(ref_arr_un[i])
                    br = brier_func(p_real_un[i], d[ref_arr_un[i]])
                    l.append(br)

            ref = np.array(ref)
            l = np.array(l)

            # user defined weights
            if weighting_method == "user_weights":

                ## compute mean
                lst_fac = np.unique(ref)
                n_fac = len(lst_fac)  # number of unit/facies to test
                weights = [dic_weights[i] for i in lst_fac]  # weights for each class, 1st order unit have a weight of 1

                res = 0
                for i in range(n_fac):
                    u_id = lst_fac[i]
                    res += weights[i]*np.mean(l[ref==u_id])

                res /= np.sum(weights)  


            # weights adjusted proportionally to the hierarchy
            elif weighting_method == "prop_weights":
                def weight_calc(unit):

                    l = []
                    def fun(unit, w):

                        if unit.mummy_unit is not None:

                            m_unit = unit.mummy_unit
                            w *= 1/len(m_unit.SubPile.list_units)
                            fun(m_unit, w)
                        else:
                            l.append(w)

                    weight = 1
                    fun(unit, weight)
                    return l[0]

                ## compute mean
                lst_fac = np.unique(ref)
                n_fac = len(lst_fac)  # number of unit/facies to test
                weights = [weight_calc(ArchTable.get_unit(ID=i, type="ID")) for i in lst_fac]  # weights for each class, 1st order unit have a weight of 1

                res = 0
                for i in range(n_fac):
                    u_id = lst_fac[i]
                    res += weights[i]*np.mean(l[ref==u_id])

                res /= np.sum(weights)  


            # same weight to each class
            elif weighting_method == "same_weights":

                ## compute mean
                lst_fac = np.unique(ref)
                n_fac = len(lst_fac)  # number of unit/facies to test
                if n_fac > 0:
                    weights = np.ones(n_fac)*1  # weights of each unit/facies

                    res = 0
                    for i in range(n_fac):
                        u_id = lst_fac[i]
                        res += weights[i]*np.mean(l[ref==u_id])

                    res /= n_fac

                else:

                    res = -1
            else:
                res = np.mean(l)

            if np.isnan(res) or res is None:
                res = -1

            scores_un.append(res)  # add score for this borehole
            #if ArchTable.verbose:
            #    print(res)

            # compute facies score
            d={}
            for i,k in enumerate(list_facies_ids):
                d[k] = i

            l = []
            ref = []
            for i in range(ref_arr_fa.shape[0]):

                if ~np.isnan(ref_arr_fa[i]):
                    ref.append(ref_arr_fa[i])
                    br = brier_func(p_real_fa[i], d[ref_arr_fa[i]])
                    l.append(br)

            ref = np.array(ref)
            l = np.array(l)

            # same weight to each class
            ## compute mean
            lst_fac = np.unique(ref)
            n_fac = len(lst_fac)  # number of unit/facies to test
            weights = np.ones(n_fac)*1  # weights of each unit/facies

            res = 0
            for i in range(n_fac):
                u_id = lst_fac[i]
                res += weights[i]*np.mean(l[ref==u_id])

            #res /= n_fac

            res = np.mean(l)
            scores_fa.append(res)

    d = {"proba_correct":(scores_correct_un, scores_correct_fa), "brier":(scores_un, scores_fa)}
    return d

# X_validation
def X_valid(ArchTable, k=3, nreal_un=5, nreal_fa=2,plot=True,
            brier = True, proba_correct = True,
            aggregate_method = None,
            weighting_method = "same_weights", dic_weights = None, parallel=False,
            seed = 15, folding_method = "random", aspect="auto",
             verbose = 1, **kwargs):

    """
    Perform a Cross-validation on the given ArchTable

    Parameters
    ----------
    ArchTable : :class:`base.Arch_table` object
        The ArchTable to perform the X-validation on
    k : int
        Number of folds
    nreal_un : int
        Number of unit realizations to estimate score
    nreal_fa : int
        Number of facies realiations to estimate score
    brier : bool
        Return brier scores
    proba_correct : bool
        Return proportion of correct cells per units/facies
    aggregate_method : str or None
        To perform X-valid on mean model rather than on realizations
        This parameter is pass to the realizations_aggregation function
        Ignored if None
    weighting_method : str
        Which method to use for applying the weights
        possible values are : same_weights, prop_weights, user_weights
    folding_method : string
        Folding method to use to separate the data
        methods availables : "random", "k_means" (to implement), "stratified" (to implement)
    plot : bool
        Display plots
    seed : int
        Seed
    verbose : 0 or 1
        Verbosity level
        
    Returns
    -------
    tuple of size 3
        (score_folds, df_conf_norm_un, df_conf_norm_fa) where
        score_folds is a list of size k containing the scores for each fold.

        Each entry of the score_folds list is a tuple containing various scores:

            - final score units
            - final score facies
            - distinct brier scores for each borehole (units and facies)
            - proportion of correct cells for each borehole (units and facies)
            - list of test boreholes
            - list of train boreholes (nreal, nreal_fa, n_boreholes)

        df_conf_norm_un is a dataframe containing the normalized confusion matrix for each unit
        df_conf_norm_fa is a dataframe containing the normalized confusion matrix for each facies
        
    """

    np.random.seed(seed)
    assert len(ArchTable.list_bhs) > 1, "there is not enough boreholes to perform K-validation"
    
    Lx = ArchTable.xg[-1] - ArchTable.xg[0]
    arch_table_dummy = copy.deepcopy(ArchTable)
    
    if nreal_fa == 0:
        extractFacies = False
    elif nreal_fa > 0:
        extractFacies = True
    else:
        extractFacies = False
        nreal_fa = 0

    if nreal_un == 0:
        extractUnits = False
    elif nreal_un > 0:
        extractUnits = True
    else:
        extractUnits = False
        nreal_un = 0

    if not extractUnits and not extractFacies:
        print("Number of units and facies realizations are equal to 0, Cross-validation aborted")
        return 

    # modify verbose
    ini_verb = arch_table_dummy.verbose
    arch_table_dummy.verbose = verbose
    arch_table_dummy.nreal_fa = nreal_fa
    
    l_bhs = arch_table_dummy.list_bhs

    # keep bhs with necessary info
    new_l_bhs = []
    if extractFacies:
        for bh in l_bhs:
            if bh.log_facies is not None:
                new_l_bhs.append(bh)

    if extractUnits:   
        for bh in l_bhs:
            if bh.log_strati is not None:
                if bh not in new_l_bhs:
                    new_l_bhs.append(bh)

    l_bhs = new_l_bhs
    n_bh = len(l_bhs)

    if k > n_bh:
        k = n_bh    

    ## Folding method --> TO DO add other methods ##
    if folding_method == "random":
            
        # determine size of a fold
        size_fold = int(n_bh/k)

        # shuffle bh ids
        ids = np.arange(n_bh)
        np.random.shuffle(ids)

    elif folding_method == "stratified":
        pass

    elif folding_method == "kmeans":
        pass

    # get folds
    folds = np.ones([k, size_fold])
    o_b = 0
    o_a = 0
    for i in range(k):

        o_a += size_fold
        folds[i] = ids[o_b:o_a]
        o_b = o_a
    
    ## 
    score_folds = []
    for ifold in range(k):

        idx_to_remove = folds[ifold]  # fold

        new_fold=[bh for i,bh in enumerate(l_bhs) if (i not in idx_to_remove)]  # train set
        bh_rm = [bh for i,bh in enumerate(l_bhs) if (i in idx_to_remove)]  # test set
        xy_fold = np.array([(i.x, i.y) for i in new_fold])  # coordinates
        xy_rm = np.array([(i.x, i.y) for i in bh_rm])

        # initialize
        arch_table_dummy.Geol = Geol()  # initialize results to nothing  
        arch_table_dummy.rem_all_bhs()  # remove previous boreholes
        arch_table_dummy.add_bh(new_fold)
        arch_table_dummy.seed = np.random.randint(1e6, 10e6)

        # reprocess
        arch_table_dummy.bhs_processed=0
        arch_table_dummy.erase_hd()
        if arch_table_dummy.geol_map is not None:
            arch_table_dummy.process_geological_map()
        arch_table_dummy.process_bhs()

        # simulations
        if parallel:
            import ArchPy.parallel
            ArchPy.parallel.parallel_compute(arch_table_dummy, nreal_un, nreal_fa, 0)
        else:
            arch_table_dummy.compute_surf(nreal_un)
            arch_table_dummy.compute_facies(nreal_fa)


        if aggregate_method is not None:  # only works on units...

            res = arch_table_dummy.realizations_aggregation(method=aggregate_method, **kwargs)

            arch_table_dummy.Geol.units_domains = res.reshape(1, arch_table_dummy.get_nz(), arch_table_dummy.get_ny(), arch_table_dummy.get_nx())
            arch_table_dummy.nreal_units= 1
            nreal_un = 1

        # extract the realizations at test bh locations
        res = np.array([(i.x, i.y) for i in bh_rm])
        pos_x = res[:, 0]
        pos_y = res[:, 1]

        # extract bh at locations
        a = arch_table_dummy.make_fake_bh(pos_x, pos_y, stratIndex=range(arch_table_dummy.nreal_units),
                              faciesIndex=range(max(1, arch_table_dummy.nreal_fa)),
                              extractFacies=extractFacies)
        a = np.array(a)

        # test this fold
        d = test_bh_1fold(arch_table_dummy, a, bh_rm, plot=plot, weighting_method=weighting_method, dic_weights=dic_weights, 
                          proba_correct = proba_correct, brier=brier, aspect=aspect)
        scores = d["brier"]
        scores = np.array(scores)

        if plot:
            plt.scatter(xy_rm[:, 0], xy_rm[:, 1])
            plt.scatter(xy_fold[:, 0], xy_fold[:, 1])
            for i in range(xy_rm.shape[0]):
                plt.text(xy_rm[i, 0], xy_rm[i, 1], s=np.round(scores[0, i], 3))
                if nreal_fa > 0:
                    plt.text(xy_rm[i, 0] - Lx/10, xy_rm[i, 1], s=np.round(scores[1, i], 3), c="r")        
            
            plt.show()
            
        final_score_un = 0
        final_score_fa = 0
        total_depth = 0
        for i in range(len(scores[0])):
            de = bh_rm[i].depth
            total_depth += de
            final_score_un += de*scores[0][i]
            final_score_fa += de*scores[1][i]
            
        final_score_un /= total_depth
        final_score_fa /= total_depth
        
        score_folds.append((final_score_un, final_score_fa, scores, np.array(d["proba_correct"]), bh_rm, a))  


    ### Begin summary ###
    print("@@@ CONFUSION MATRIX: UNITS @@@ \n")

    res = score_folds.copy()

    y_actu = []
    y_pred = []
    y_actu_fa = []
    y_pred_fa = []

    if nreal_fa > 0:
        fa_flag = True
    else:
        fa_flag = False

    n_bh = len(res[0][-2])
    for ik in range(k):
        for ifa in range(max(1, nreal_fa)):
            for ibh in range(n_bh):
                bh_reals = res[ik][-1][:, ifa, ibh]
                bh_true = res[ik][-2][ibh]

                if fa_flag:
                    a = bh2array(ArchTable, bh_true, typ=None)
                    arr_true = a[0]
                    arr_true_fa = a[1]

                    mask_fa = arr_true_fa == arr_true_fa
                else:
                    arr_true = bh2array(ArchTable, bh_true)
                mask = arr_true == arr_true

                for ireal in range(nreal_un):

                    if fa_flag:
                        a = bh2array(ArchTable, bh_reals[ireal], typ=None)
                        arr_real = a[0]
                        arr_real_fa = a[1]

                        for i in range(len(arr_true_fa[mask_fa])):
                            v_fa = arr_real_fa[mask][i]
                            v2_fa = arr_true_fa[mask][i]

                            if v_fa == v_fa and v2_fa == v2_fa:
                                y_actu_fa.append(v2_fa)
                                y_pred_fa.append(v_fa)

                    else:
                        arr_real = bh2array(ArchTable, bh_reals[ireal])

                    for i in range(len(arr_true[mask])):
                        v = arr_real[mask][i]
                        v2 = arr_true[mask][i]

                        if v == v and v2 == v2:
                            y_actu.append(v2)
                            y_pred.append(v)

    df_conf_norm_un = None
    df_conf_norm_fa = None

    # matrix units     
    y_actu = np.array(y_actu)
    y_pred = np.array(y_pred)

    y_actu = pd.Series(y_actu, name='Actual')
    y_pred = pd.Series(y_pred, name='Predicted')
    df_confusion = pd.crosstab(y_pred, y_actu, rownames=['Predicted'], colnames=['Actual'])

    for ival in df_confusion.index:
        if ival not in df_confusion.columns:
            df_confusion[ival] = np.ones([df_confusion.shape[0]])*np.nan

    for ival in df_confusion.columns:
        if ival not in df_confusion.index:
            df_confusion.loc[ival] = np.ones([df_confusion.shape[1]])*np.nan

    df_conf_norm = df_confusion.div(df_confusion.sum(axis=1), axis="index")
    df_conf_norm.columns = [ArchTable.get_unit(ID=i, type="ID").name[:] for i in df_conf_norm.columns]
    df_conf_norm.columns.name = df_confusion.columns.name
    df_conf_norm.index = [ArchTable.get_unit(ID=i, type="ID").name[:] for i in df_conf_norm.index]
    df_conf_norm.index.name = df_confusion.index.name

    figsize = (int(df_conf_norm.shape[0]/3), int(df_conf_norm.shape[0]/3))
    plt.figure(figsize=figsize)
    plot_confusion_matrix(df_conf_norm, title="Normalized Confusion matrix")
    df_conf_norm_un = df_conf_norm.copy()

    #matrix facies
    if fa_flag:
        print("@@@ CONFUSION MATRIX: FACIES @@@ \n")
        y_actu_fa = np.array(y_actu_fa)
        y_pred_fa = np.array(y_pred_fa)

        y_actu = pd.Series(y_actu_fa, name='Actual')
        y_pred = pd.Series(y_pred_fa, name='Predicted')
        df_confusion = pd.crosstab(y_actu, y_pred, rownames=['Actual'], colnames=['Predicted'])
        df_conf_norm = df_confusion.div(df_confusion.sum(axis=1), axis="index")
        df_conf_norm.columns = [ArchTable.get_facies_obj(ID=i, type="ID").name[:] for i in df_conf_norm.columns]
        df_conf_norm.columns.name = df_confusion.columns.name
        df_conf_norm.index = [ArchTable.get_facies_obj(ID=i, type="ID").name[:] for i in df_conf_norm.index]
        df_conf_norm.index.name = df_confusion.index.name
        figsize = (int(df_conf_norm.shape[0]/3), int(df_conf_norm.shape[0]/3))
        plt.figure(figsize=figsize)
        plot_confusion_matrix(df_conf_norm, title="Normalized Confusion matrix")
        df_conf_norm_fa = df_conf_norm.copy()

# other things to add ?

    return (np.array(score_folds), df_conf_norm_un, df_conf_norm_fa)

def plot_confusion_matrix(df, title='Confusion matrix', cmap="plasma"):

    plt.imshow(df, cmap=cmap) # imshow
    plt.title("Confusion matrix")
    plt.colorbar()
    tick_marks = np.arange(len(df.columns))
    plt.xticks(tick_marks, df.columns, rotation=90)
    plt.yticks(tick_marks, df.index)
    #plt.tight_layout()
    plt.ylabel(df.index.name)
    plt.xlabel(df.columns.name)
    plt.show()


import sys, platform
import setuptools

with open("README.md", "r") as file:
    long_desc = file.read()

setuptools.setup(
    name='Geoarchpy',
    version='1.0.0',
    author="Ludovic Schorpp",
    author_email="ludovic.schorpp@unine.ch",
    description="Simpler geological and property models",
    long_description=long_desc,
    long_description_content_type='text/markdown',
    url = "https://github.com/randlab/ArchPy",
    install_requires=['matplotlib',
                      'numpy',
                      'pyvista==0.37.0',
                      'scipy',
                      'numba',
                      'scikit-learn',
                      'pyyaml',
                      'pandas',
                      'shapely < 2.0',
                       'rasterio',
                       'geopandas'],  # download geone
    packages=setuptools.find_packages(),
    include_package_data=True,
    #data_files=[("lib\\site-packages\\ArchPy\\libraries", ["ArchPy\\libraries\\cov_facies.dll"])],
    license=open('LICENSE', encoding='utf-8').read()
)
